//****************************************************************************
//
// File         : camera.cpp
//
// Description  : Camera control functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <math.h>
#include <string.h>

#include "main.h"
#include "matrix.h"
#include "camera.h"
#include "input.h"
#include "vector.h"

#ifdef AUDIO

#include "\c\audio\include\audio.h"

extern HAC         voice1;
extern LPAUDIOWAVE roket_sound;

#endif

Vector_3D   speed;

//--- Set camera matrix -----------------------------------------------------

void Set_Camera(Camera *cam)
{
    memcpy(&(engine.camera), cam, sizeof(Camera));
}

//--- Reset camera interface ------------------------------------------------

void Init_Interface(void)
{
    engine.ticks = 0;
    speed.x = 0;
    speed.y = 0;
    speed.z = 0;

    Update_Input();
    Reset_Timer();
}

//--- Examine input and update camera, rocket mode --------------------------

void Update_Rocket(void)
{
    float           lookspeed, lookspeedx, lookspeedy, movespeed;
    float           sx, cx, sxx, cxx, sxy, cxy, step;
    Matrix_4x4      xmatrix, ymatrix, zmatrix;

    Mat_Identity_4x4(xmatrix);
    Mat_Identity_4x4(ymatrix);
    Mat_Identity_4x4(zmatrix);
    
    step = (double)engine.ticks / TICKS_PER_FRAME;

    lookspeed = engine.camera.lookspeed * step;
    lookspeedx = -(float)engine.camera.mousespeed * mousedx;
    lookspeedy = (float)engine.camera.mousespeed * mousedy;

    sxx = sin(lookspeedx);
    cxx = cos(lookspeedx);

    sxy = sin(lookspeedy);
    cxy = cos(lookspeedy);

    // Look left/rigth

    ymatrix[0][0] = cxx;
    ymatrix[2][0] = -sxx;
    ymatrix[0][2] = sxx;
    ymatrix[2][2] = cxx;

    xmatrix[1][1] = cxy;
    xmatrix[2][1] = sxy;
    xmatrix[1][2] = -sxy;
    xmatrix[2][2] = cxy;

    sx = sin(lookspeed);
    cx = cos(lookspeed);

    if (KEY(LEFT_KEY)) {

    	zmatrix[0][0] = cx;
       	zmatrix[1][0] = -sx;
       	zmatrix[0][1] = sx;
       	zmatrix[1][1] = cx;
    }

    if (KEY(RIGHT_KEY)) {

       	zmatrix[0][0] = cx;
       	zmatrix[1][0] = sx;
       	zmatrix[0][1] = -sx;
       	zmatrix[1][1] = cx;
    }

    // Move forwards/backwards

    movespeed = engine.camera.acceleration * step;

    if (mousebuttons & LEFT_MBUTTON) {
    	speed.x += movespeed * engine.camera.cmatrix[2][0];
        speed.y += movespeed * engine.camera.cmatrix[2][1];
        speed.z += movespeed * engine.camera.cmatrix[2][2];


        #ifdef AUDIO
            AStartVoice(voice1);
            ASetVoiceFrequency(voice1, 1000 + (int)(Vector_Magnitude(&speed) * 30));
        #endif
    } else {
        #ifdef AUDIO
            AStopVoice(voice1);
            ASetVoicePosition(voice1, 0);
        #endif
    }
    speed.y -= engine.camera.gravity * step;

    speed.x *= (1 - (step * engine.camera.friction));
    speed.y *= (1 - (step * engine.camera.friction));
    speed.z *= (1 - (step * engine.camera.friction));

    Mat_Concatenate_4x4(engine.camera.cmatrix, xmatrix);
    Mat_Concatenate_4x4(engine.camera.cmatrix, ymatrix);
    Mat_Concatenate_4x4(engine.camera.cmatrix, zmatrix);

    engine.camera.location.x += step * speed.x;
    engine.camera.location.y += step * speed.y;
    engine.camera.location.z += step * speed.z;
}

//--- Read keyboard and update camera, Descent mode -------------------------

void Update_Ship(void)
{
    float           lookspeed, movespeed;
    float           sx, cx, step;
    Matrix_4x4      xmatrix, ymatrix, zmatrix;

    Mat_Identity_4x4(xmatrix);
    Mat_Identity_4x4(ymatrix);
    Mat_Identity_4x4(zmatrix);

    step = (double)engine.ticks / TICKS_PER_FRAME;

    movespeed = 10 * step;
    lookspeed = engine.camera.lookspeed * step;

    sx = sin(lookspeed);
    cx = cos(lookspeed);

     // Roll left/right

    if (KEY(Q_KEY)) {

        if (!KEY(RSHIFT_KEY)) {
            zmatrix[0][0] = cx;
            zmatrix[1][0] = -sx;
            zmatrix[0][1] = sx;
            zmatrix[1][1] = cx;
        }
        else {
            zmatrix[0][0] = cx;
            zmatrix[1][0] = sx;
            zmatrix[0][1] = -sx;
            zmatrix[1][1] = cx;
        }
    }
 
    if (!KEY(LALT_KEY)) {

        // Look left/rigth

        if (KEY(LEFT_KEY)) {
             ymatrix[0][0] = cx;
             ymatrix[2][0] = -sx;
             ymatrix[0][2] = sx;
             ymatrix[2][2] = cx;
         }

         if (KEY(RIGHT_KEY)) {
             ymatrix[0][0] = cx;
             ymatrix[2][0] = sx;
             ymatrix[0][2] = -sx;
             ymatrix[2][2] = cx;
         }

         // Look up/down

        if (KEY(UP_KEY)) {
           xmatrix[1][1] = cx;
           xmatrix[2][1] = -sx;
           xmatrix[1][2] = sx;
           xmatrix[2][2] = cx;
        }

        if (KEY(DOWN_KEY)) {
           xmatrix[1][1] = cx;
           xmatrix[2][1] = sx;
           xmatrix[1][2] = -sx;
           xmatrix[2][2] = cx;
        }
    } else {

        // Strafe left/right

        if (KEY(LEFT_KEY)) {
            engine.camera.location.x -= movespeed * engine.camera.cmatrix[0][0];
            engine.camera.location.y -= movespeed * engine.camera.cmatrix[0][1];
            engine.camera.location.z -= movespeed * engine.camera.cmatrix[0][2];
        }
        if (KEY(RIGHT_KEY)) {
            engine.camera.location.x += movespeed * engine.camera.cmatrix[0][0];
            engine.camera.location.y += movespeed * engine.camera.cmatrix[0][1];
            engine.camera.location.z += movespeed * engine.camera.cmatrix[0][2];
        }

        // Strafe up/down

        if (KEY(UP_KEY)) {
            engine.camera.location.x -= movespeed * engine.camera.cmatrix[1][0];
            engine.camera.location.y -= movespeed * engine.camera.cmatrix[1][1];
            engine.camera.location.z -= movespeed * engine.camera.cmatrix[1][2];
        }
        if (KEY(DOWN_KEY)) {
            engine.camera.location.x += movespeed * engine.camera.cmatrix[1][0];
            engine.camera.location.y += movespeed * engine.camera.cmatrix[1][1];
            engine.camera.location.z += movespeed * engine.camera.cmatrix[1][2];
        }
    }

     // Move forward/backward

    if (KEY(A_KEY)) {
        engine.camera.location.x += movespeed * engine.camera.cmatrix[2][0];
        engine.camera.location.y += movespeed * engine.camera.cmatrix[2][1];
        engine.camera.location.z += movespeed * engine.camera.cmatrix[2][2];
    }
    if (KEY(Z_KEY)) {
        engine.camera.location.x -= movespeed * engine.camera.cmatrix[2][0];
        engine.camera.location.y -= movespeed * engine.camera.cmatrix[2][1];
        engine.camera.location.z -= movespeed * engine.camera.cmatrix[2][2];
    }

    Mat_Concatenate_4x4(engine.camera.cmatrix, xmatrix);
    Mat_Concatenate_4x4(engine.camera.cmatrix, ymatrix);
    Mat_Concatenate_4x4(engine.camera.cmatrix, zmatrix);
}
