//****************************************************************************
//
// File         : console.cpp
//
// Description  : Command console functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "main.h"
#include "input.h"
#include "text.h"
#include "memory.h"
#include "video.h"
#include "console.h"

#define HISTORY_SIZE    65536
#define HISTORY_PREV    0
#define HISTORY_NEXT    1
#define HISTORY_COLOR   2
#define HISTORY_FLAGS   3

#define CBACK           0x3c3c3c3c
#define WHITE           0xfb
#define RED             245
#define LGREY           243
#define DGREY           244

void    Redraw_Cline(char *o, char *n);
char    *NextCmd(char *cmd);
char    *PrevCmd(char *cmd);

int     cvisible, mode_flags;
int     ctextx, ctexty;
char    *history, *console_start, *history_end;

//--- Initialize the console ------------------------------------------------

int Init_Console(void)
{
    if (NULL == (history = (char *)malloc(HISTORY_SIZE))) return(FALSE);
    memset(history, 0, HISTORY_SIZE);
    cvisible = FALSE;

    console_start = history;
    history_end = history + history[HISTORY_NEXT];
    ctexty = 4;
    CPutStr("-------------------------", LGREY);
    CPutStr("  THE R3D CONSOLE", WHITE);
    CPutStr("-------------------------", LGREY);
    return(TRUE);
}

//--- Free console history buffer -------------------------------------------
                                               
void Close_Console(void)
{
    free(history);
}

//--- Enter console ---------------------------------------------------------

int Console(void)
{
    char    input[64], done = FALSE;
    int     returncode;

    returncode = 0;

    // No clearing, no pageflipping

    mode_flags = engine.screen.currentmode->flags;
    Set_PageFlip(FALSE);
    engine.screen.currentmode->flags &= ~MODEINFO_CLEAR;
    cvisible = TRUE;

    Render_Frame();
    Redraw_Console();

    do {
        CGetStr(input, 235);
        if (!strcmp(input, "QUIT")) {
            returncode = QUIT;
            break;
        }
        if (!strcmp(input, "EXIT")) {
            done = TRUE;
            continue;
        }

        if (strlen(input)) Process_Command(input);
        if ((KEY(ESC_KEY)) || (KEY(TAB_KEY))) done = TRUE;

    } while (!done);
    while ((KEY(ESC_KEY)) || (KEY(TAB_KEY))) Update_Input();

    cvisible = FALSE;

    // Restore screenmode

    engine.screen.currentmode->flags = (short)mode_flags;
    if (mode_flags & MODEINFO_PAGEFLIP) Set_PageFlip(TRUE);
    else Set_PageFlip(FALSE);
    Render_Frame();

    Reset_Timer();
    return(returncode);
}

//--- Execute console command -----------------------------------------------

void Process_Command(char *str)
{
    char    *command, *args, output[64], valid = FALSE;
    int     i;

    command = str;
    while (*command < 33) command ++;
    args = command + strlen(command) - 1;
    while (*args < 33) *args-- = 0;

    args = NULL;
    for (i = 1; i < (int)strlen(command); i++) {
        if (command[i] == 32) {
            args = command + i;
            *args++ = 0;
            break;
        }
    }

	if (!strcmp(command, "VER")) {
		strcpy(output, R3DVERSION);
        StrUppercase(output);
		CPutStr(output, LGREY);
		valid = TRUE;
	}

    if (!strcmp(command, "LEVEL")) {
        sprintf(output, "SECTORS  : %d", engine.nr_sectors);
        CPutStr(output, LGREY);
        sprintf(output, "PORTALS  : %d", engine.nr_portals);
        CPutStr(output, LGREY);
        sprintf(output, "POLYGONS : %d", engine.nr_polygons);
        CPutStr(output, LGREY);
        valid = TRUE;
    }

    if (!strcmp(command, "POV")) {
        sprintf(output, "POV IS %.1f,%.1f,%.1f", engine.camera.location.x,
            engine.camera.location.y, engine.camera.location.z);
        CPutStr(output, LGREY);
        valid = TRUE;
    }

    if (!strcmp(command, "FOV")) {
        if (args) {
            float fov;
            sscanf(args, "%f", &fov);
            engine.screen.pmulx = fov;
            engine.screen.pmuly = fov / engine.screen.pmul_ratio;
        }
        sprintf(output, "FOV IS %.1f", engine.screen.pmulx);
        CPutStr(output, LGREY);
        Render_Frame();
        Redraw_Console();
        valid = TRUE;
    }

    if (!strcmp(command, "ACCEL")) {
        if (args) {
            float accel;
            sscanf(args, "%f", &accel);
            engine.camera.acceleration = accel * 0.01;
        }
        sprintf(output, "ACCELERATION IS %.2f", engine.camera.acceleration * 100);
        CPutStr(output, LGREY);
        valid = TRUE;
    }

    if (!strcmp(command, "GRAVITY")) {
        if (args) {
            float gravity;
            sscanf(args, "%f", &gravity);
            engine.camera.gravity = gravity * 0.01;
        }
        sprintf(output, "GRAVITY IS %.2f", engine.camera.gravity * 100);
        CPutStr(output, LGREY);
        valid = TRUE;
    }

    if (!strcmp(command, "MSPEED")) {
        if (args) {
            float mspeed;
            sscanf(args, "%f", &mspeed);
            engine.camera.mousespeed = mspeed * 0.001;
        }
        sprintf(output, "MOUSESPEED IS %.2f", engine.camera.mousespeed * 1000);
        CPutStr(output, LGREY);
        valid = TRUE;
    }

    if (!strcmp(command, "FRICTION")) {
        if (args) {
            float friction;
            sscanf(args, "%f", &friction);
            engine.camera.friction = friction * 0.01;
        }
        sprintf(output, "FRICTION IS %.2f", engine.camera.friction * 100);
        CPutStr(output, LGREY);
        valid = TRUE;
    }

    if (!strcmp(command, "GFXCARD")) {
        sprintf(output, "CARDNAME    : %s %s", engine.screen.vendorname,
            engine.screen.cardname);
        StrUppercase(output);
        CPutStr(output, LGREY);
        StrUppercase(output);
        sprintf(output, "USABLE VMEM : %dK", engine.screen.videomemory >> 10);
        CPutStr(output, LGREY);
        valid = TRUE;
    }

    if (!strcmp(command, "MODE")) {
        if (args) {

            int     xres, yres, set;

            sscanf(args, "%dX%d", &xres, &yres);

            set = Set_ScreenMode(xres, yres);

            if (!set) {
                sprintf(output, "CAN'T SET VIDEOMODE %s", args);
                CPutStr(output, RED);
            } else {
                mode_flags = engine.screen.currentmode->flags;
                Set_PageFlip(FALSE);
                engine.screen.currentmode->flags &= ~MODEINFO_CLEAR;
            }
 

            Render_Frame();
            Redraw_Console();
        }
        else {
            char *msg = "VGA";
            sprintf(output, "RESOLUTION    : %s %dX%d", msg, engine.screen.xres, engine.screen.yres);
            CPutStr(output, LGREY);

            sprintf(output, "WINDOW        : %dX%d", engine.screen.width, engine.screen.height);
            CPutStr(output, LGREY);

            if (mode_flags & MODEINFO_CLEAR) msg = "YES"; else msg = "NO";
            sprintf(output, "CLEARING      : %s", msg);
            CPutStr(output, LGREY);

            if (mode_flags & MODEINFO_PAGEFLIP) msg = "YES"; else msg = "NO";
            sprintf(output, "PAGE FLIPPING : %s", msg);
            CPutStr(output, LGREY);

            if (mode_flags & MODEINFO_WVBL) msg = "YES"; else msg = "NO";
            sprintf(output, "WAITVBL       : %s", msg);
            CPutStr(output, LGREY);
        }
        valid = TRUE;
    }

    if (!strcmp(command, "MODELIST")) {
        for (i = 0; i < engine.screen.nr_modes; i++) {
            sprintf(output, "%dX%d", engine.screen.modelist[i].xres,
                engine.screen.modelist[i].yres);
			CPutStr(output, LGREY);
		}
		valid = TRUE;
	}	    

    if (!strcmp(command, "CLEAR")) {
        if (mode_flags & MODEINFO_CLEAR) {
            CPutStr("CLEAR OFF", LGREY);
            mode_flags &= ~MODEINFO_CLEAR;
        } else {
            CPutStr("CLEAR ON", LGREY);
            mode_flags |= MODEINFO_CLEAR;
        }
        valid = TRUE;
    }

    if (!strcmp(command, "TMAP")) {
        if (engine.flags & ENGINE_TMAPPING) {
            CPutStr("TEXTUREMAPPING OFF", LGREY);
            engine.flags &= ~ENGINE_TMAPPING;
        } else {
            CPutStr("TEXTUREMAPPING ON", LGREY);
            engine.flags |= ENGINE_TMAPPING;
        }
        Render_Frame();
        Redraw_Console();
        valid = TRUE;
    }


    if (!strcmp(command, "WVBL")) {
        if (mode_flags & MODEINFO_WVBL) {
            CPutStr("WAITVBL OFF", LGREY);
            mode_flags &= ~MODEINFO_WVBL;
        } else {
            CPutStr("WAITVBL ON", LGREY);
            mode_flags |= MODEINFO_WVBL;
        }
        valid = TRUE;
    }

    if (!strcmp(command, "PFLIP")) {
        if (mode_flags & MODEINFO_PAGEFLIP) {
            CPutStr("PAGEFLIPPING OFF", LGREY);
            mode_flags &= ~MODEINFO_PAGEFLIP;
        } else {
                if (!(engine.screen.currentmode->flags & MODEINFO_FLIPABLE)) {
                CPutStr("NOT SUPPORTED IN THIS MODE", RED);
            } else {
                CPutStr("PAGEFLIPPING ON", LGREY);
                mode_flags |= MODEINFO_PAGEFLIP;
            }
        }
        valid = TRUE;
    }

    if (!strcmp(command, "SAVETGA")) {
        Render_Frame();
        Save_TGA();
        CPutStr("TGA SCREENSHOT SAVED...", LGREY);
        Redraw_Console();
        valid = TRUE;
    }

    if (!valid) CPutStr("UNKNOWN COMMAND", RED);
}

//--- Redraw the console ----------------------------------------------------

void Redraw_Console(void)
{
    char    len;
    char    *str = console_start;

    ctexty = 4;

    Dword_Fill(engine.screen.buffer + engine.screen.width, CBACK, (engine.screen.width * 98) >> 2);
    Dword_Fill(engine.screen.buffer, 0xfbfbfbfb, engine.screen.width >> 2);
    Dword_Fill(engine.screen.buffer + engine.screen.width * 99, 0xfbfbfbfb, engine.screen.width >> 2);

    for (int i = 0; i < 11; i++) {
        len = str[HISTORY_NEXT];
        if (!len) break;
        Set_TextColor(str[HISTORY_COLOR], 0);
        PutStringXY(4, ctexty, str + 4);
        ctexty += 8;
        str += len;
    }
    engine.screen.Show();
}

//--- Output string to console ----------------------------------------------

void CPutStr(char *str, char color)
{
    if (*str == 0) return;
    if (cvisible) {
        Set_TextColor(color, 0);
        PutStringXY(4, ctexty, str);
    }
    int slen = strlen(str) + 5;
    history_end[HISTORY_NEXT] = (char)slen;
    history_end[HISTORY_COLOR] = color;
    history_end[HISTORY_FLAGS] = 0;
    strcpy(history_end + 4, str);
    history_end += history_end[HISTORY_NEXT];
    history_end[HISTORY_PREV] = (char)slen;
    CNewLine();
}

//--- Input string from console ---------------------------------------------

void CGetStr(char *str, char color)
{
    char    ch, up, down, *strstart, *oldcmd, *inpos, scrolled = FALSE;

    ctextx = 11;
    strstart = str;
    oldcmd = history_end;
    inpos = console_start;

    *strstart = 0;

    Set_TextColor(color, 0);
    PutStringXY(4, ctexty, ":");
    Cursor(ctextx, ctexty);
    engine.screen.Show();

    while (13 != (ch = GetChar())) {

        if ((KEY(ESC_KEY)) || (KEY(TAB_KEY))) {
            console_start = inpos;
            *strstart = 0;
            return;
        }

       if (KEY(PGUP_KEY)) {
            console_start -= console_start[HISTORY_PREV];
            Redraw_Console();
            scrolled = TRUE;
        }
        if (KEY(PGDN_KEY)) {
            if (console_start != inpos) {
                console_start += console_start[HISTORY_NEXT];
                Redraw_Console();
                scrolled = TRUE;
            }
        }

        if (KEY(UP_KEY)) up = TRUE; else up = FALSE;
        if (KEY(DOWN_KEY)) down = TRUE; else down = FALSE;

        if ((ch && scrolled) || (scrolled && (console_start == inpos)) || up || down) {
            console_start = inpos;
            Redraw_Console();
            scrolled = FALSE;
            Set_TextColor(color, 0);
            PutCharXY(4, ctexty, ':');
            PutStringXY(11, ctexty, strstart);
            Cursor(ctextx, ctexty);
        }

        if (up) {
            oldcmd = PrevCmd(oldcmd);
            char *laststr = oldcmd + 5;
            Redraw_Cline(strstart, laststr);
            strcpy(strstart, laststr);
            str = strstart + strlen(laststr);
        }
 
        if (down) {
            oldcmd = NextCmd(oldcmd);
            char *laststr = oldcmd + 5;
            Redraw_Cline(strstart, laststr);
            strcpy(strstart, laststr);
            str = strstart + strlen(laststr);
        }
 
        if (ch) {
            if (ch == 8) {
                if (ctextx <= 11) continue;
                BlankOut(ctextx, ctexty);
                ctextx -= 7;
                BlankOut(ctextx, ctexty);
                str--;
            } else if (ctextx < 310) {
                *str++ = ch;
                BlankOut(ctextx, ctexty);
                PutCharXY(ctextx, ctexty, ch);
                ctextx += 7;
            }
            *str = 0;
            Cursor(ctextx, ctexty);
        }
        engine.screen.Show();
    }

    int slen = strlen(strstart) + 6;
    history_end[HISTORY_NEXT] = (char)slen;
    history_end[HISTORY_COLOR] = color;
    history_end[HISTORY_FLAGS] = 1;
    history_end[4] = ':';
    strcpy(history_end + 5, strstart);
    history_end += history_end[HISTORY_NEXT];
    history_end[HISTORY_PREV] = (char)slen;

    BlankOut(ctextx, ctexty);
    if (scrolled) {
        console_start = inpos;
        Redraw_Console();
        ctexty -= 8;
    }
    CNewLine();
}

//--- Redraw cursor line ----------------------------------------------------

void Redraw_Cline(char *o, char *n)
{
    ctextx = 11;

    do {
        BlankOut(ctextx, ctexty);
        ctextx += 7;
    } while (*o++);

    ctextx = 11;
    while (*n) {
        PutCharXY(ctextx, ctexty, *n);
        ctextx += 7;
        n++;
    }
    Cursor(ctextx, ctexty);
}

//--- Find previos command --------------------------------------------------

char *PrevCmd(char *cmd)
{
    char *first = cmd;
    do {
        if (cmd[HISTORY_PREV] == 0) return(first);
        else cmd -= cmd[HISTORY_PREV];
    } while (cmd[HISTORY_FLAGS] != 1);
    return(cmd);
}

//--- Find next command -----------------------------------------------------

char *NextCmd(char *cmd)
{
    do {
        if (cmd[HISTORY_NEXT] == 0) return(cmd);
        else cmd += cmd[HISTORY_NEXT];
    } while (cmd[HISTORY_FLAGS] != 1);
    return(cmd);
}

//--- Step to next line, scroll up if necessary -----------------------------

void CNewLine()
{
    if (cvisible) engine.screen.Show();
    ctexty += 8;
    if (ctexty > 84) {
        console_start += console_start[HISTORY_NEXT];
        if (cvisible) Redraw_Console();
        ctexty = 84;
    }
}

//--- Clear character -------------------------------------------------------

void BlankOut(int x, int y)
{
    char    *dest = engine.screen.buffer + y * engine.screen.width + x;

    for (int i = 0; i < 8; i++) {
        *((int *)dest) = CBACK;
        *((int *)dest + 1) = CBACK;
        dest += engine.screen.width;
    }
}

//--- Draw cursor -----------------------------------------------------------

void Cursor(int x, int y)
{
    char    *dest = engine.screen.buffer + (y + 5) * engine.screen.width + x;

    *((int *)dest) = 0xfbfbfb00;
    *((int *)dest + 1) = 0xfbfbfbfb;
    dest += engine.screen.width;
    *((int *)dest) = 0;
    *((short *)dest + 2) = 0;
    *((char *)dest + 6) = 0;
}

