//****************************************************************************
//
// File         : video.cpp
//
// Description  : DOS/DJGPP low level videomode functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <dpmi.h>
#include <dos.h>
#include <go32.h>
#include <pc.h>
#include <sys/movedata.h>
#define NDEBUG
#include <assert.h>
#include <sys/nearptr.h>

#include "main.h"
#include "video.h"
#include "tweaked.h"
#include "vesa2.h"

void    Detect_SVGA(void);
int     Set_SVGAMode(int mode);

char            svga_databuffer[2024];
short           oldmode = 0, modelist[64];
char            *pbp;

int             tweaked_rowlen, tweaked_ysize;
int             mode_rowlen[5] = {80,80,90,90,90};
int             mode_ysize[5] = {240,400,270,360,400};

ModeInfo        standard_dosmodes[6] = {{320, 200, 0, 0, MODEINFO_VGA},
                    {320, 240, 0, 1, MODEINFO_XVGA},
                    {320, 400, 0, 2, MODEINFO_XVGA},
                    {360, 270, 0, 3, MODEINFO_XVGA},
                    {360, 360, 0, 4, MODEINFO_XVGA},
                    {360, 400, 0, 5, MODEINFO_XVGA}};

//--- Init video system ------------------------------------------------------

int Init_Video(void)
{
    union REGS regs;

    regs.h.ah = 0xf;
    int86(0x10, &regs, &regs);
    oldmode = (short)regs.h.al;

    engine.screen.lfb = NULL;
    engine.screen.vendorname = "UNKNOWN";    // Assume standard 128K VGA card
    engine.screen.cardname = "STANDARD VGA";
    engine.screen.videomemory = 131072;
    
    engine.screen.modelist = (ModeInfo *)malloc(6 * sizeof(ModeInfo));
    if (!engine.screen.modelist) return(FALSE);

    // Copy Standard VGA modes...

    memcpy(engine.screen.modelist, standard_dosmodes, 6 * sizeof(ModeInfo));
    engine.screen.nr_modes = 6;

    Detect_SVGA();

    return(TRUE);
}

//--- Close down video system -----------------------------------------------

void Close_Video()
{
    union REGS  regs;

    if (engine.screen.modelist) free(engine.screen.modelist);

    if (oldmode) {
        regs.w.ax = oldmode;
        int86(0x10, &regs, &regs);
    }
}

//--- Set video mode --------------------------------------------------------

int Set_VideoMode(int mode)
{
    int         i;
    Register    *regs;
    union REGS  cpuregs;

    __djgpp_nearptr_enable();

    if (mode == 0) {
        engine.screen.Show = Copy_Buffer;
        engine.screen.BClear = Clear_Buffer;
        engine.screen.VClear = Clear_VGA;

        cpuregs.w.ax = 0x13;
        int86(0x10, &cpuregs, &cpuregs);
        engine.screen.lfb = (char *)(0xa0000 + __djgpp_conventional_base);
        return(TRUE);
    }

    if (mode < 6) {

        engine.screen.Show = Copy_XVGA;
        engine.screen.BClear = Clear_Buffer;
        engine.screen.VClear = Clear_XVGA;

        switch(mode) {
            case 1 : regs = Mode_320x240regs; break;
            case 2 : regs = Mode_320x400regs; break;
            case 3 : regs = Mode_360x270regs; break;
            case 4 : regs = Mode_360x360regs; break;
            case 5 : regs = Mode_360x400regs; break;
        }

        tweaked_rowlen = mode_rowlen[mode-1];    // Bytes per row
        tweaked_ysize = mode_ysize[mode-1];
        cpuregs.d.eax = 0x13;
        int86(0x10, &cpuregs, &cpuregs);      // Set normal mode 13h

        Wait_Retrace();
        outp(0x3d4, 0x11);                     // Enable writes to registers 0-7
        i = inp(0x3d5) & 0x7F;
        outp(0x3d4, 0x11);
        outp(0x3d5, i);

        i = 25;
        while (i--) OutReg(*regs++);           // Output VGA registers

        engine.screen.lfb = (char *)(0xa0000 + __djgpp_conventional_base);
        return(TRUE);
    }

    engine.screen.Show = Copy_Buffer;
    engine.screen.BClear = Clear_Buffer;
    engine.screen.VClear = Clear_VGA;

    if (Set_SVGAMode(engine.screen.modelist[mode].code)) return(TRUE);
    else return(FALSE);

}

//--- Look for VESA 2.0 video card ------------------------------------------

void Detect_SVGA(void)
{
    int            modes;
    VBEInfo        vbeinfo;
    VBEModeInfo    vbemode;
    ModeInfo       *modeinfo;
    __dpmi_regs    r;
    char           *buffer;
    short          *vbemodelist;

    assert(sizeof(VBEInfo) < _go32_info_block.size_of_transfer_buffer);
    vbeinfo.VbeSignature = VBE2;             // VBE2 = 'VBE2'
    r.x.ax = 0x4f00;
    r.x.di = __tb & 0x0f;
    r.x.es = (__tb >> 4) & 0xffff;
    dosmemput(&vbeinfo, sizeof(VBEInfo), __tb);
    if (__dpmi_int(0x10, &r)) return;
    dosmemget(__tb, sizeof(VBEInfo), &vbeinfo);

    if (r.x.ax == 0x4f) {
        if ((vbeinfo.VbeSignature == VESA) && (vbeinfo.VbeVersion >= 0x200)) {

            engine.screen.videomemory = vbeinfo.TotalMemory << 16;

            // Copy strings to protected mode memory

            __djgpp_nearptr_enable();

            buffer = svga_databuffer;
            
            strcpy(buffer, (char *)FLAT_PTR(vbeinfo.OemVendorNamePtr));
            engine.screen.vendorname = buffer;
            buffer += 1 + strlen(buffer);

            strcpy(buffer, (char *)FLAT_PTR(vbeinfo.OemProductNamePtr));
            engine.screen.cardname = buffer;
            buffer += 1 + strlen(buffer);

            // Copy video mode list to protected mode memory

            vbemodelist = (short *)FLAT_PTR(vbeinfo.VideoModePtr);

            modes = 0;
            while(*vbemodelist != -1)
                modelist[modes++] = *vbemodelist++;

            modeinfo = &(engine.screen.modelist[engine.screen.nr_modes]);
            realloc(engine.screen.modelist, (engine.screen.nr_modes + modes) * sizeof(ModeInfo));

            __djgpp_nearptr_disable();

            // Add modes to mode list

            for (int i = 0; i < modes; i++) {

                // Registers passed to the real mode interuppt

                r.x.ax = 0x4f01;          // Function 0x4f01
                r.x.cx = modelist[i];     // Mode number
                r.x.di = __tb & 0x0f;
                r.x.es = (__tb >> 4) & 0xffff;
 
                // Simulate real mode interrupt

                __dpmi_int(0x10, &r);
                dosmemget(__tb, sizeof(VBEModeInfo), &vbemode);

                // Ok, Now extract useful modes

                if ((vbemode.ModeAttributes & VBEMODE_CAPS_LINEAR) &&
                    (vbemode.BitsPerPixel == 8)) {

                    modeinfo->code = modelist[i];
                    modeinfo->xres = vbemode.XResolution;
                    modeinfo->yres = vbemode.YResolution;
                    modeinfo->size = 0;
                    modeinfo->flags = MODEINFO_SVGA;

                    if (engine.screen.videomemory >= (modeinfo->xres * modeinfo->yres * 2))
                        modeinfo->flags |= MODEINFO_FLIPABLE;
                    pbp = (char *)(vbemode.PhysBasePtr);

                    modeinfo++;
                    engine.screen.nr_modes++;
                }
            }
        }
    }
}

//--- Set VESA 2.0 linear framebuffer mode ----------------------------------

int Set_SVGAMode(int mode)
{
    __dpmi_regs     r;
    __dpmi_meminfo  mi;

    r.x.ax = 0x4f02;                          // Function 0x4f02
    r.x.bx = (short)(mode | VBE_MODEFLAG_LFB | VBE_MODEFLAG_NOCLEAR);
    __dpmi_int(0x10, &r);

    if (r.x.ax != 0x4f) {
        Close_VideoMode();
        return(FALSE);
    }

    // Map frame buffer into reachable memory using DJGPP DPMI function

    mi.size = 1024*1024*4;                      //  4 Megs
    mi.address = (unsigned long)pbp;
    __dpmi_physical_address_mapping(&mi);
    engine.screen.lfb = (char *)mi.address + __djgpp_conventional_base;

    return(TRUE);
}

//--- Close down current video mode -----------------------------------------

void Close_VideoMode()
{
    __dpmi_meminfo  mi;

    if (engine.screen.lfb) {
        mi.address = (unsigned long)engine.screen.lfb;
        __dpmi_free_physical_address_mapping(&mi);

        engine.screen.lfb = NULL;
    }
}

//--- Copy linear frame buffer to video memory ------------------------------

void Copy_Buffer(void)
{
    char        *source, *dest;

    source = engine.screen.buffer;
    dest = engine.screen.videomem +
        ((engine.screen.yres - engine.screen.height) >> 1) * engine.screen.xres +
        ((engine.screen.xres - engine.screen.width) >> 1);

    int dwords = engine.screen.width >> 2;

    if (engine.screen.flags & MODEINFO_WVBL) Wait_Retrace();

    if (engine.screen.flags & MODEINFO_CLEAR) {
        for (int i = 0; i < engine.screen.height; i++) {
            Dword_CopyClear(dest, source, 0, dwords);
            source += engine.screen.width;
            dest += engine.screen.xres;
        }
    } else {
        for (int i = 0; i < engine.screen.height; i++) {
            Dword_Copy(dest, source, dwords);
            source += engine.screen.width;
            dest += engine.screen.xres;
        }
    }
}

//--- Clear linear frame buffer ---------------------------------------------

void Clear_Buffer(void)
{
    char *dest = engine.screen.buffer +
        (int)(engine.screen.top_clip) * engine.screen.xres +
        (int)(engine.screen.left_clip);

    for (int i = 0; i < engine.screen.height; i++) {
        Dword_Fill(dest, 0, engine.screen.width >> 2);
        dest += engine.screen.xres;
    }

}

void Clear_VGA(void)
{
    if (engine.screen.currentmode->flags & MODEINFO_FLIPABLE)
        Dword_Fill(engine.screen.lfb, 0, (engine.screen.xres * engine.screen.yres * 2) >> 2);
    else Dword_Fill(engine.screen.videomem, 0, (engine.screen.xres * engine.screen.yres) >> 2);
}


//--- Copy buffer to chained video memory -----------------------------------

void Copy_XVGA(void)
{
    int x = (engine.screen.xres - engine.screen.width) >> 1;
    int y = (engine.screen.yres - engine.screen.height) >> 1;

    Tweaked_CopyBlock(x, y, engine.screen.width,
        engine.screen.height, engine.screen.buffer);

    if (engine.screen.flags & MODEINFO_CLEAR)
        Dword_Fill(engine.screen.buffer, 0, (engine.screen.width * engine.screen.height) >> 2);
}

void Clear_XVGA(void)
{
    Tweaked_ClearBlock(0, 0, engine.screen.xres, engine.screen.yres, 0);
}


//--- Wait for vertical retrace ---------------------------------------------

void Wait_Retrace(void)
{
    asm(" movw  $0x3da, %%dx    \n\t"
        " 1:                    \n\t"
        " inb   %%dx, %%al      \n\t"
        " test  $8, %%al        \n\t"
        " jnz   1b              \n\t"
        " 2:                    \n\t"
        " inb   %%dx, %%al      \n\t"
        " test  $8, %%al        \n\t"
        " jz    2b                  "
        :
        :
        : "%eax", "%edx");
}

//--- Set VGA palette -------------------------------------------------------

void Set_Palette(char *palette)
{
    asm(" cld                   \n\t"
        " movl  $0x3c8, %%edx   \n\t"
        " xorl  %%eax, %%eax    \n\t"
        " outb  %%al, %%dx      \n\t"
        " incl  %%edx           \n\t"
        " movl  $768, %%ecx     \n\t"
        " rep                   \n\t"
        " outsb                 \n\t"
        :
        : "S" (palette)
        : "%eax", "%ecx", "%edx", "%esi");
}

//--- Sets VESA 2.0 scanline width ------------------------------------------

int Set_ScanWidth(long width)
{

    union REGS regs;

    regs.x.ax = 0x4f06;
    regs.x.bx = 0;
    regs.x.cx = width;
    int86(0x10, &regs, &regs);
    if (regs.x.ax!=0x4f) return(FALSE);

    return(TRUE);
}

//--- Sets VESA 2.0 display start location ----------------------------------

int Set_DisplayStart(long x, long y)
{
    union REGS regs;

    regs.x.ax=0x4f07;
    regs.x.bx=0;
    regs.x.cx=x;
    regs.x.dx=y;
    int86(0x10, &regs, &regs);
    if (regs.x.ax!=0x4f) return(FALSE);

    return(TRUE);
}

//--- Output VGA register. Modified from tweak16b example source ------------

void OutReg(Register r)
{
    switch (r.port) {
        /* First handle special cases: */

        case ATTRCON_ADDR:
            inp(STATUS_ADDR);           /* reset read/write flip-flop */
            outp(ATTRCON_ADDR, r.index | 0x20);
                                  /* ensure VGA output is enabled */
            outp(ATTRCON_ADDR, r.value);
        break;

        case MISC_ADDR:
        case VGAENABLE_ADDR:
            outp(r.port, r.value);      /* directly to the port */
        break;

        case SEQ_ADDR:
        case GRACON_ADDR:
        case CRTC_ADDR:
        default:                       /* This is the default method: */
            outp(r.port, r.index);      /* index to port           */
            outp(r.port+1, r.value);    /* value to port+1         */
        break;
    }
}

//--- Clear block on chained video memory ------------------------------------

void ClearBlockAsm(int xsize, int ysize, int color, char *dest);

void ClearBlockAsm(int xsize, int ysize, int color, char *dest)
{
    asm(" pushl %%edx           \n\t"
        " movw  $0x3c4, %%dx    \n\t"
        " movl  %%eax, %%ecx    \n\t"
        " movl  $0xf02, %%eax   \n\t"
        " outw  %%ax, %%dx      \n\t"
        " movb  %%bl, %%bh      \n\t"
        " movl  %%ebx, %%eax    \n\t"
        " shll  $16, %%ebx      \n\t"
        " popl   %%edx          \n\t"
        " movw  %%ax, %%bx      \n\t"
        " movl  %%ecx, %%eax    \n\t"
        " movl   _tweaked_rowlen, %%esi \n\t"
        " addl  %%eax, %%eax    \n\t"
        " subl  %%eax, %%esi    \n\t"
        " xorb  %%ah, %%ah      \n\t"
        " shrl  $1, %%ecx       \n\t"
        " jnc   1f              \n\t"
        " movb  $1, %%ah        \n\t"
    "1:                         \n\t"       // YLoop
        " movb  %%cl, %%al      \n\t"
    "2:                         \n\t"       // XLoop
        " movl  %%ebx, (%%edi)  \n\t"
        " addl  $4, %%edi       \n\t"
        " decb  %%al            \n\t"
        " jnz   2b              \n\t"
        " orb   %%ah, %%ah      \n\t"
        " jz    3f              \n\t"
        " movw  %%bx, (%%edi)   \n\t"
        " addl  $2, %%edi       \n\t"
    "3:                         \n\t"       // Even
        " addl  %%esi, %%edi    \n\t"
        " dec   %%edx           \n\t"
        " jnz   1b               "
        :
        : "a" (xsize), "d" (ysize), "b" (color), "D" (dest)
        : "%eax", "%ebx", "%ecx", "%edx", "%edi", "%esi");
}

void Tweaked_ClearBlock(int x, int y, int xsize, int ysize, int color)
{
    ClearBlockAsm(xsize >> 3, ysize, color, engine.screen.videomem +
        y * tweaked_rowlen + (x >> 2));
}

//--- Copy block to chained video memory ------------------------------------

void CopyBlockAsm(int xsize, int ysize, char *dest, char *source);

void CopyBlockAsm(int xsize, int ysize, char *dest, char *source)
{
    asm(" movl  %%eax, %%ebp    \n\t"
        " movl  %%eax, %%ecx    \n\t"
        " shrl  $1, %%eax       \n\t"
        " shll  $3, %%ebp       \n\t"
        " movb  %%al, %%ch      \n\t"
        " subl  $4, %%ebp       \n\t"
    "1:                         \n\t"       // YLoop
        " xorb  %%cl, %%cl      \n\t"
    "2:                         \n\t"       // PLoop
        " movl  $0x3c4, %%edx   \n\t"
        " movl  $0x102, %%eax   \n\t"
        " shlb  %%cl, %%ah      \n\t"
        " pushl %%edi           \n\t"
        " outw  %%ax, %%dx      \n\t"
        " pushl %%esi           \n\t"
        " movb  %%ch, %%dl      \n\t"
     "3:                        \n\t"       // XLoop
        " movb  12(%%esi), %%ah \n\t"
        " movb  8(%%esi), %%al  \n\t"
        " shll  $16, %%eax      \n\t"
        " movb  4(%%esi), %%ah  \n\t"
        " movb  (%%esi), %%al   \n\t"
        "                       \n\t"
        " movl  %%eax, (%%edi)  \n\t"
        " addl  $16, %%esi      \n\t"
        " addl  $4, %%edi       \n\t"
        " decb  %%dl            \n\t"
        " jnz   3b              \n\t"
        "                       \n\t"
        " popl  %%esi           \n\t"
        " incb  %%cl            \n\t"
        " incl  %%esi           \n\t"
        " popl  %%edi           \n\t"
        " cmpb  $4, %%cl        \n\t"
        " jne   2b              \n\t"
        "                       \n\t"
        " addl  _tweaked_rowlen, %%edi \n\t"
        " addl  %%ebp, %%esi    \n\t"
        "                       \n\t"
        " decl  %%ebx           \n\t"
        " jnz   1b                  "
        :
        : "a" (xsize), "b" (ysize), "D" (dest), "S" (source)
        : "%eax", "%ebx", "%ecx", "%edx", "%edi", "%esi", "%ebp");
}

void Tweaked_CopyBlock(int x, int y, int xsize, int ysize, char *source)
{
    CopyBlockAsm(xsize >> 3, ysize, engine.screen.videomem +
        y * tweaked_rowlen + (x >> 2), source);
}


