//****************************************************************************
//
// File         : engine.cpp
//
// Description  : Main portal engine functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <string.h>
#include <float.h>

#include "main.h"
#include "poly.h"
#include "polygon.h"
#include "clip.h"
#include "fixed.h"
#include "text.h"
#include "timer.h"
#include "input.h"
#include "console.h"

inline void Transform_Vertice(Vertice_3D *v1, Vertice_3D *v2);

Engine          engine;                         // Global engine structure
Vertice_2DI     xy[VERTICES_PER_POLYGON * 2];   // Temporary vertice data

//--- Setup and shut down... ------------------------------------------------

int Init_Engine(void)
{
    // If using DJGPP, turn off float->int overflow execption...

    #ifdef COMP_DJGPP
        _control87(EM_INVALID, EM_INVALID);
    #endif

    if (!Init_Mouse()) return(FALSE);
    memset(&engine, 0, sizeof(Engine));
    if (!Init_Console()) return(FALSE);
    if (!Init_Screen()) return(FALSE);
    return (TRUE);
}

void Close_Engine(void)
{
    Close_Mouse();
    Close_Console();
    Close_Screen();
}

//--- Render world at currect camera position -------------------------------

void Render_Frame(void)
{
    Update_CSector();                           // Find current sector

    // Setup initial clipping values...

    Sector *sector_p = engine.csector_p;        // Current sector
    Reset_Clip(sector_p);
    engine.screen.cclip = &(sector_p->clip);

    Draw_Sector(sector_p);                      // Draw sectors, reqursively

    // Measure frame time and display FPS

    if (engine.framecounter) engine.ticks = Elapsed_Ticks();
    else engine.ticks = 1;

    Reset_Timer();
    engine.total_ticks += engine.ticks;
    engine.framecounter++;

    int xoffs = (int)engine.screen.left_clip;
    int yoffs = (int)engine.screen.top_clip;

    Set_TextColor(237, 0);
    PutStringXY(xoffs + 4, yoffs + 2, "FPS:");
    Set_TextColor(235, 0);
    PutIntXY(xoffs + 36, yoffs + 2, TICKS_PER_SECOND / engine.ticks);
}

//--- Find the current sector -----------------------------------------------

void Update_CSector(void)
{
    Polygon *poly = &(engine.csector_p->polygons_p[0]);

    float x = engine.camera.location.x + (1 * engine.camera.cmatrix[2][0]);
    float y = engine.camera.location.y + (1 * engine.camera.cmatrix[2][1]);
    float z = engine.camera.location.z + (1 * engine.camera.cmatrix[2][2]);

    for (int p = 0; p < engine.csector_p->nr_polygons; p++) {

        if (poly->flags & POLY_PORTAL) {

            float dist = x * poly->normal.x +
                         y * poly->normal.y +
                         z * poly->normal.z + poly->d;

            if (dist < 0) {
                engine.csector_p = (Sector *)poly->tplane_p;
                return;
            }
        }
        poly++;
    }
}

//--- Draw sectors recursively ----------------------------------------------

void Draw_Sector(Sector *sector)
{
    int         i, nr_portals, corners;
    Clip        *tmpclip;

    nr_portals = 0;
    tmpclip = engine.screen.cclip;
    engine.pVertices = sector->vertices_p;
    engine.screen.cclip = &(sector->clip);

    Transform_Sector(sector);
    Project_Sector(sector);

    for (int p = 0; p < sector->nr_polygons; p++) {

        Polygon *poly = &(sector->polygons_p[p]);

        corners = Setup_FlatPolygon(poly, xy);
        if (!corners) continue;

        if (poly->flags & POLY_PORTAL) {
            for (int i = 0; i < corners; i++)
                sector->portals[nr_portals].xy[i] = xy[i];

            sector->portals[nr_portals].sector_p = poly->tplane_p;
            sector->portals[nr_portals].nr_vertices = (short)corners;
            nr_portals++;
        }
        else {
            if (engine.flags & ENGINE_TMAPPING) {
                Setup_TPlane(poly->tplane_p);
                Draw_SPTextPolygon(xy, corners, poly->tplane_p);
            }
            else {
                Draw_FlatPolygon(xy, corners, poly->attribute);
            }
        }
    }

    for (i = 0; i < nr_portals; i++) {
        Portal *portal = &(sector->portals[i]);
        if (Setup_Portal(portal)) {
            Draw_Sector((Sector *)(portal->sector_p));
        }
    }

    engine.screen.cclip = tmpclip;          // Restore clipdata pointer
}

//--- Setup texture plane ---------------------------------------------------

void Setup_TPlane(TPlane *tp)
{
    Vector_3D   p, m, n;

    p = tp->p;

    m.x = tp->m.x * engine.camera.cmatrix[0][0] + tp->m.y * engine.camera.cmatrix[0][1] + tp->m.z * engine.camera.cmatrix[0][2];
    m.y = tp->m.x * engine.camera.cmatrix[1][0] + tp->m.y * engine.camera.cmatrix[1][1] + tp->m.z * engine.camera.cmatrix[1][2];
    m.z = tp->m.x * engine.camera.cmatrix[2][0] + tp->m.y * engine.camera.cmatrix[2][1] + tp->m.z * engine.camera.cmatrix[2][2];

    n.x = tp->n.x * engine.camera.cmatrix[0][0] + tp->n.y * engine.camera.cmatrix[0][1] + tp->n.z * engine.camera.cmatrix[0][2];
    n.y = tp->n.x * engine.camera.cmatrix[1][0] + tp->n.y * engine.camera.cmatrix[1][1] + tp->n.z * engine.camera.cmatrix[1][2];
    n.z = tp->n.x * engine.camera.cmatrix[2][0] + tp->n.y * engine.camera.cmatrix[2][1] + tp->n.z * engine.camera.cmatrix[2][2];

    Transform_Vertice(&p, &p);

    Cross_Product(&p, &n, &(tp->a));
    Cross_Product(&m, &p, &(tp->b));
    Cross_Product(&n, &m, &(tp->c));

    tp->a.x *= 127;
    tp->a.y *= 127;
    tp->a.z *= 127;

    tp->b.x *= 127;
    tp->b.y *= 127;
    tp->b.z *= 127;

}

//--- Transform sector vertices ---------------------------------------------

inline void Transform_Sector(Sector *sector)
{
    for (int v = 0; v < sector->nr_vertices; v++) {
        Transform_Vertice(&(engine.transformed_p[v]),
            &(sector->vertices_p[v]));
    }
}

inline void Transform_Vertice(Vertice_3D *v1, Vertice_3D *v2)
{
    Vertice_3D tmp;
    tmp.x = v2->x - engine.camera.location.x;
    tmp.y = v2->y - engine.camera.location.y;
    tmp.z = v2->z - engine.camera.location.z;

    v1->x = tmp.x * engine.camera.cmatrix[0][0] + tmp.y * engine.camera.cmatrix[0][1] + tmp.z * engine.camera.cmatrix[0][2];
    v1->y = tmp.x * engine.camera.cmatrix[1][0] + tmp.y * engine.camera.cmatrix[1][1] + tmp.z * engine.camera.cmatrix[1][2];
    v1->z = tmp.x * engine.camera.cmatrix[2][0] + tmp.y * engine.camera.cmatrix[2][1] + tmp.z * engine.camera.cmatrix[2][2];
}

//--- Project sector vertices -----------------------------------------------

inline void Project_Sector(Sector *sector)
{

    double xcenter = (double)engine.screen.xcenter;
    double ycenter = (double)engine.screen.ycenter;

    for (int v = 0; v < sector->nr_vertices; v++) {

        double zinv = (double)1 / (engine.transformed_p[v].z);

        engine.projected_p[v].x = xcenter +
            engine.transformed_p[v].x * engine.screen.pmulx * zinv;

        engine.projected_p[v].y = ycenter +
            engine.transformed_p[v].y * engine.screen.pmuly * zinv;
    }
}
