//****************************************************************************
//
// File         : polygon.cpp
//
// Description  : Polygon/Portal setup functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include "main.h"

#include "polygon.h"
#include "clip.h"
#include "fixed.h"

//--- Setup flat polygon ----------------------------------------------------

int Setup_FlatPolygon (Polygon *poly, Vertice_2DI *xy)
{
    int             i, z, corners;
    Vertice_2DD     xyf[VERTICES_PER_POLYGON * 2];
    Vertice_3D      u;

    // Count number of pixels behind NEAR_CLIP

    for (i = z = 0; i < poly->nr_vertices; i++)
        if (engine.transformed_p[poly->pindex[i]].z <= (NEAR_CLIP)) z++;

    // Totally behind NEAR_CLIP?

    if (z == poly->nr_vertices) return(0);

    if (!z) {
        corners = poly->nr_vertices;
        for (i = 0; i < poly->nr_vertices; i++)
            xyf[i] = engine.projected_p[poly->pindex[i]];
    }
    else corners = ZClip_FlatPolygon (xyf, poly);

    if (corners < 3) return(0);

    // Do backface culling...

    u.x = engine.camera.location.x;
    u.y = engine.camera.location.y;
    u.z = engine.camera.location.z;

    u.x -= engine.pVertices[poly->pindex[0]].x;
    u.y -= engine.pVertices[poly->pindex[0]].y;
    u.z -= engine.pVertices[poly->pindex[0]].z;

    if (0 >= Dot_Product(&u, &(poly->normal))) return(0);

    // Clip polygon to the portals' bounding box...

    corners = WClip_FlatPolygon (xyf, corners);
    if (corners < 3) return(0);

    // Convert coordinates to fixed point...

    for (i = 0; i < corners; i++) {
        xy[i].x = Double2Int (xyf[i].x * FIXED_ONE);
        xy[i].y = Double2Int (xyf[i].y * FIXED_ONE);
    }

    return(corners);
}

//--- Setup portal polygon --------------------------------------------------

int Setup_Portal(Portal *portal)
{
    int             v, i, y, scry, ydif, minx, maxx, miny, maxy;
    int             points, lx, rx, ldx, rdx, lcount, rcount;
    int             top_index, bottom_index, ledge, redge;

    Vertice_2DI     *xy;
    Clip            *cclip, *sclip;

    sclip = &(((Sector *)portal->sector_p)->clip);
    cclip = engine.screen.cclip;
    xy = portal->xy;
    points = portal->nr_vertices;

    // Find top and bottom points

    minx = maxx = xy[0].x;
    miny = maxy = xy[0].y;
    top_index = bottom_index = 0;

    for (i = 1; i < points; i++) {
       if (xy[i].x < minx)
          minx = xy[i].x;

       if (xy[i].x > maxx)
          maxx = xy[i].x;
 
       if (xy[i].y < miny) {
          miny = xy[i].y;
          top_index = i;
       }
       if (xy[i].y > maxy) {
          maxy = xy[i].y;
          bottom_index = i;
       }
    }

    if (maxy == miny) return(FALSE);

    minx = F2I_Ceil(minx);
    maxx = F2I_Ceil(maxx);
    miny = F2I_Ceil(miny);
    maxy = F2I_Ceil(maxy);

    sclip->top_clip = miny;
    sclip->left_clip = minx;
    sclip->right_clip = maxx;
    sclip->bottom_clip = maxy;

    ledge = redge = top_index;
    lcount = rcount = 0;
    scry = miny;

    v = FALSE;

    // Loop forever

    while (1) {

       // New left edge?

        if (--lcount < 0) {

            if (ledge == bottom_index) return(v);
            i = ledge + 1;
            if (i == points) i = 0;
            lcount = F2I_Ceil(xy[i].y - FIXED_ONE) - F2I_Ceil(xy[ledge].y);

            while (lcount < 0) {
                if (i == bottom_index) return(v);
                ledge = i;
                i++;
                if (i == points) i = 0;
                lcount = F2I_Ceil(xy[i].y - FIXED_ONE) - F2I_Ceil(xy[ledge].y);
            }

            // Calculate edge gradients

            y = (xy[ledge].y + (FIXED_ONE - 1)) & ~FIXED_MASK;
            ydif = xy[i].y - xy[ledge].y;
            if (ydif < (FIXED_ONE / 2)) ydif = (FIXED_ONE / 2);

            ldx = Fixed_Div64(xy[i].x - xy[ledge].x, ydif);
            lx = xy[ledge].x;

            lx += Fixed_Mul64(y - xy[ledge].y, ldx);

            ledge = i;
        }

        // New right edge?

        if (--rcount < 0) {

            if (redge == bottom_index) return(v);
            i = redge-1;
            if (i < 0) i = points - 1;

            rcount = F2I_Ceil(xy[i].y - FIXED_ONE) - F2I_Ceil(xy[redge].y);

            while (rcount < 0) {
                if (i == bottom_index) return(v);
                redge = i;
                i--;
                if (i < 0) i = points - 1;
                rcount = F2I_Ceil(xy[i].y - FIXED_ONE) - F2I_Ceil(xy[redge].y);
            }

            // Calculate edge gradients

            y = (xy[redge].y + (FIXED_ONE - 1)) & ~FIXED_MASK;
            ydif = xy[i].y - xy[redge].y;
            if (ydif < (FIXED_ONE / 2)) ydif = (FIXED_ONE / 2);

            rdx = Fixed_Div64(xy[i].x - xy[redge].x, ydif);
            rx = xy[redge].x;

            rx += Fixed_Mul64(y - xy[redge].y, rdx);

            redge = i;
        }

        if (!((lx > cclip->hclip[scry].ex) || (rx < cclip->hclip[scry].sx)))
            v = TRUE;

        if (lx < cclip->hclip[scry].sx)
            sclip->hclip[scry].sx = cclip->hclip[scry].sx;
        else
            sclip->hclip[scry].sx = lx;

        if (rx > cclip->hclip[scry].ex)
            sclip->hclip[scry].ex = cclip->hclip[scry].ex;
        else
            sclip->hclip[scry].ex = rx;

        lx += ldx;
        rx += rdx;
        scry++;
    }
}
