//****************************************************************************
//
// File         : pwd.cpp
//
// Description  : Functions to load and initialize a PWD file.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "main.h"
#include "memory.h"
#include "pwd.h"

#define TEXTURE_SIZE 128

void        Init_Polys(void);
Sector      *Find_Sector(int s);

int         max_nr_vertices;
FILE        *pwdfile;

//--- Read world data from PWD file -----------------------------------------

int Load_PWDWorld(char *filename)
{
    int                 i;

    PWD_Header          pwdheader;
    PWD_TextureHeader   textheader;

    engine.nr_sectors = engine.nr_polygons = engine.nr_portals = 0;
    max_nr_vertices = 0;

    pwdfile = fopen(filename, "rb");
    if (!pwdfile) {
        puts("Load_PWDFile() : Can't open pwdfile.");
        return(FALSE);
    }

    fread(&pwdheader, sizeof(PWD_Header), 1, pwdfile);

    engine.objects_p = (Object *)AllocMem(pwdheader.nr_objects *
        (sizeof(Object) + sizeof(int)));

    if (!engine.objects_p) {
        puts("Load_PWDFile() : Not enough memory");
        return(FALSE);
    }

    for (i = 0; i < pwdheader.nr_objects; i++) {
        if (!(Read_Object(&(engine.objects_p[i])))) {
            return(FALSE);
        }
    }

    engine.csector_p = &(engine.objects_p[0].sectors_p[0]);
    engine.nr_objects = pwdheader.nr_objects;
    engine.nr_textures = pwdheader.nr_textures;

    engine.transformed_p = (Vertice_3D *)AllocMem(max_nr_vertices *
        (sizeof(Vertice_2DD) + sizeof(Vertice_3D)) + 768);

    if (!engine.transformed_p) {
        puts("Load_PWDFile() : Not enough memory");
        return(FALSE);
    }

    engine.projected_p = (Vertice_2DD *)(engine.transformed_p + max_nr_vertices);
    engine.palette_p = (char *)(engine.projected_p + max_nr_vertices);

    printf("Objects  : %d\nTextures : %d\n", engine.nr_objects, engine.nr_textures);

    // Load textures

    if (engine.nr_textures) {

        if (NULL == (engine.textures_p = (char *)AllocMem(TEXTURE_SIZE *
                TEXTURE_SIZE * (engine.nr_textures + 1)))) {
            puts("Load_PWDFile() : Not enough memory 2");
            return(FALSE);
        }
        printf ("Memory allocated : %d bytes\n", TEXTURE_SIZE * TEXTURE_SIZE * engine.nr_textures);

        char    *loadptr, *dest = engine.textures_p;
        int     textures_left = engine.nr_textures;

        while (textures_left) {
            for (int x = 0; x < (256 / TEXTURE_SIZE); x++) {
                fread(&textheader, sizeof(PWD_TextureHeader), 1, pwdfile);
                loadptr = dest;
                for (int y = 0; y < textheader.ysize; y++) {
                    fread(loadptr, textheader.xsize, 1, pwdfile);
                    loadptr += 256;
                }
                dest += TEXTURE_SIZE;
                textures_left--;
                if (!textures_left) break;
            }
            dest = dest + 256 * (TEXTURE_SIZE - 1);
        }
    }

    // Load palette

    fseek(pwdfile, -768, SEEK_END);
    fread(engine.palette_p, 768, 1, pwdfile);
/*
    for (i = 0; i < 256; i++) {
        engine.palette_p[i+i+i+1] >>= 1;
        engine.palette_p[i+i+i+2] >>= 1;
    }
*/

    fclose(pwdfile);
    Init_Polys();

    return(TRUE);
}


//--- Initialize polygons ---------------------------------------------------

void Init_Polys(void)
{
    Vector_3D       v1, v2, normal;

    for (int o = 0; o < engine.nr_objects; o++) {

        Sector *sector = engine.objects_p[o].sectors_p;
        engine.nr_sectors += engine.objects_p[o].nr_sectors;

        for (int s = 0; s < engine.objects_p[o].nr_sectors; s++) {

            Polygon *poly = sector->polygons_p;
            engine.nr_polygons += sector->nr_polygons;

            for (int p = 0; p < sector->nr_polygons; p++) {

                TPlane *tplane = &(sector->tplanes_p[p]);

                Make_Vector_3D(&(sector->vertices_p[poly->pindex[1]]),
                    &(sector->vertices_p[poly->pindex[0]]), &v1);

                Make_Vector_3D(&(sector->vertices_p[poly->pindex[3]]),
                    &(sector->vertices_p[poly->pindex[0]]), &v2);

                Cross_Product(&v1, &v2, &normal);
                Normalize_Vector(&normal);
                poly->normal = normal;

                poly->d = -normal.x * sector->vertices_p[poly->pindex[0]].x -
                           normal.y * sector->vertices_p[poly->pindex[0]].y -
                           normal.z * sector->vertices_p[poly->pindex[0]].z;

                if (poly->flags & POLY_PORTAL) {
                    poly->tplane_p = (TPlane *)Find_Sector(poly->attribute);
                    poly->attribute = 1;
                    engine.nr_portals++;
                }
                else {
//                    poly->flags |= POLY_TEXTUREMAPPED;
//                    printf("%d\n", poly->attribute);

                    tplane->p = (Vector_3D)sector->vertices_p[poly->pindex[0]];

//                    Normalize_Vector(&v1);
//                    Normalize_Vector(&v2);
                    tplane->n = v1;
                    tplane->m = v2;

                    poly->attribute = 15 - poly->attribute;
                    tplane->texture_p = engine.textures_p +
                        (poly->attribute & 1) * 128 + (poly->attribute >> 1) * 256 * 128;

                    poly->tplane_p = tplane;

                    poly->attribute = (p + 39) & 255;
                    tplane->timestamp = poly->attribute;
                }

                poly++;

            }
            sector++;
        }
    }
}

//--- Find sector -----------------------------------------------------------

Sector *Find_Sector(int s)
{
    int object_id = s >> 16;
    int sector_id = s & 0xffff;

    Object *object = engine.objects_p;

    for (int o = 0; o < engine.nr_objects; o++) {
        if (object->id == object_id) {

            Sector *sector = object->sectors_p;

            for (int s = 0; s < object->nr_sectors; s++) {
                if (sector->id == sector_id) return(sector);
                sector++;
            }
        }
        object++;
    }
    return(NULL);
}

//--- Read object data from PWD file ----------------------------------------

int Read_Object(Object *object_p)
{
    int                 i;
    PWD_ObjectHeader    objheader;
    PWD_PolyHeader      polyheader;

    fread(&objheader, sizeof(PWD_ObjectHeader), 1, pwdfile);

    if (objheader.nr_vertices > max_nr_vertices)
        max_nr_vertices = objheader.nr_vertices;

    object_p->sectors_p = (Sector *)AllocMem(objheader.nr_sectors *
        sizeof(Sector) + objheader.nr_vertices * sizeof(Vertice_3D) +
        objheader.nr_polygons * sizeof(Polygon));

    if (!(object_p->sectors_p)) {
        puts("Read_Object() : Not enough memory");
        return(FALSE);
    }

    object_p->vertices_p = (Vertice_3D *)(object_p->sectors_p +
        objheader.nr_sectors);
    object_p->polygons_p = (Polygon *)(object_p->vertices_p +
        objheader.nr_vertices);

    for (i = 0; i < objheader.nr_sectors; i++) {
        if (!(Read_Sector(&(object_p->sectors_p[i])))) {
            return(FALSE);
        }
    }

    // Copy Object info

    object_p->id = objheader.id;
    object_p->nr_vertices = objheader.nr_vertices;
    object_p->nr_polygons = objheader.nr_polygons;
    object_p->nr_sectors = objheader.nr_sectors;
    object_p->flags = objheader.flags;

    // Read object vertices/polygons

    fread(object_p->vertices_p, objheader.nr_vertices, sizeof(PWD_Vertex), pwdfile);

    for (i = 0; i < objheader.nr_polygons; i++) {
        fread(&polyheader, sizeof(PWD_PolyHeader), 1, pwdfile);
        fread(object_p->polygons_p[i].pindex, polyheader.points, sizeof(short), pwdfile);
    }

    return(TRUE);
}

//--- Read sector data from PWD file ----------------------------------------

inline int Read_Sector(Sector *sector_p)
{

    PWD_SectorHeader    sectorheader;
    PWD_PolyHeader      polyheader;

    fread(&sectorheader, sizeof(PWD_SectorHeader), 1, pwdfile);

    if (sectorheader.nr_vertices > max_nr_vertices)
        max_nr_vertices = sectorheader.nr_vertices;

    sector_p->vertices_p = (Vertice_3D *)AllocMem(sectorheader.nr_vertices *
        (sizeof(Vertice_3D) + sizeof(TPlane)) + sectorheader.nr_polygons *
        sizeof(Polygon));

    if (!(sector_p->vertices_p)) {
        puts("Read_Sector() : Not enough memory");
        return(FALSE);
    }

    sector_p->polygons_p = (Polygon *)(sector_p->vertices_p +
        sectorheader.nr_vertices);
    sector_p->tplanes_p = (TPlane *)(sector_p->polygons_p +
        sectorheader.nr_polygons);

    // Copy Sector info

    sector_p->id = sectorheader.id;
    sector_p->nr_vertices = sectorheader.nr_vertices;
    sector_p->nr_polygons = sectorheader.nr_polygons;
    sector_p->flags = sectorheader.flags;

    /* Read sector vertices/polygons */

    fread(sector_p->vertices_p, sectorheader.nr_vertices, sizeof(PWD_Vertex), pwdfile);

    for (int i = 0; i < sectorheader.nr_polygons; i++) {
        fread(&polyheader, sizeof(PWD_PolyHeader), 1, pwdfile);
        sector_p->polygons_p[i].nr_vertices = polyheader.points;
        sector_p->polygons_p[i].flags = polyheader.flags;
        sector_p->polygons_p[i].attribute = polyheader.attribute;

//        if (!(polyheader.flags & POLY_PORTAL)) {
//            if (polyheader.attribute == -1)
//                sector_p->polygons_p[i].attribute = 0;

//             printf("Attrib : %d\n", sector_p->polygons_p[i].attribute);
//         }

//            sector_p->polygons_p[i].attribute=15-polyheader.attribute;

        fread(sector_p->polygons_p[i].pindex, polyheader.points, sizeof(short), pwdfile);
    }

//    printf("Vertices : %d\nPolygons : %d\n", sector_p->nr_vertices, sector_p->nr_polygons);

    return(TRUE);
}

//--- Free all allocated memory ---------------------------------------------

void Free_World(void)
{
    FreeAll();
}
