//****************************************************************************
//
// File         : roketz.cpp
//
// Description  : Roketz 3D Engine 0.1a main source file.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <stdio.h>
#include <stdlib.h>

#include "main.h"
#include "pwd.h"
#include "input.h"
#include "console.h"
#include "screen.h"
#include "timer.h"

#ifdef AUDIO

#define FREQ(nPeriod) (((LONG) console_sound->nSampleRate * 428) / nPeriod)

#include "\c\audio\include\audio.h"

AUDIOINFO       info;
LPAUDIOMODULE   lpModule;
LPAUDIOWAVE     console_sound, roket_sound;
HAC             voice1;

#endif

void    Play(void);
void    Stop(char *message, int code);

//--- Main ------------------------------------------------------------------

int main(int, char *argc[])
{
    puts(STARTUPMESSAGE);

    if (!Init_Engine())
        Stop("Bye...", 1);

    if (!Load_PWDWorld(argc[1]))
        Stop("Error! Can't load world.", 1);

    if (!(Set_ScreenMode(320, 200)))
        Stop("Error! Can't set video mode.", 1);

	Init_Timer();
	Init_Keyboard();

#ifdef AUDIO
    AInitialize();
    /* open audio device */
    info.nDeviceId = AUDIO_DEVICE_MAPPER;
    info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_STEREO;
    info.nSampleRate = 44100;

    #ifdef USEFILTER
    /* enable antialias dynamic filtering */
    info.wFormat |= AUDIO_FORMAT_FILTER;
    #endif

    AOpenAudio(&info);

    /* load module file */
    ALoadModuleFile("nocorner.mod", &lpModule, 0);
    ALoadWaveFile("blip.wav", &console_sound, 0);
    ALoadWaveFile("rocket.wav", &roket_sound, 0);
    roket_sound->wFormat |= AUDIO_FORMAT_LOOP;
    roket_sound->dwLoopStart = 0;
    roket_sound->dwLoopEnd = roket_sound->dwLength-1;


    /* open voices and play module */
    AOpenVoices(lpModule->nTracks + 1);

//    APlayModule(lpModule);
    ASetModuleVolume(16);

   /* play the waveform through a voice */
    ACreateAudioVoice(&voice1);

    ASetVoiceVolume(voice1, 24);
    ASetVoicePanning(voice1, 128);
#endif

    Play();

#ifdef AUDIO
    /* stop module and close voices */
    AStopModule();
    AStopVoice(voice1);
    ADestroyAudioVoice(voice1);

    ACloseVoices();

    /* release module file */
    AFreeModuleFile(lpModule);
    AFreeWaveFile(console_sound);

    /* close audio device */
    ACloseAudio();
#endif

    Stop("", 0);

    return(0);
}

//--- Test engine -----------------------------------------------------------

void Play(void)
{
    char        roket = FALSE, done = FALSE;
    Camera      player;

    Init_Interface();
    player.movespeed = 0;
    player.strafespeed = 12;
    player.lookspeed = DEG2RAD(2);
    player.mousespeed = 0.005;    

    player.acceleration = 0.50;
    player.gravity = 0.03;
    player.friction = 0.0025;

    Mat_Identity_4x4(player.cmatrix);

    player.location.x = 0;
    player.location.y = 0;
    player.location.z = -300;

    player.cmatrix[0][3] = 0;           // Initial position
    player.cmatrix[1][3] = 0;
    player.cmatrix[2][3] = -300;

    player.cmatrix[0][0] = -1;
    player.cmatrix[1][1] = -1;

    Set_Camera(&player);
    engine.flags |= ENGINE_TMAPPING;

    while (!(KEY(ESC_KEY) || done)) {

	    Update_Input();
	
	    if (roket) Update_Rocket(); else Update_Ship();

        if (TestKey(F1_KEY)) {
            if (!roket) {
                roket = TRUE;
                CPutStr("ROCKET MODE ACTIVATED", 243);
                #ifdef AUDIO
                    APlayVoice(voice1, roket_sound);
                    AStopVoice(voice1);
                #endif
            }
            else {
                roket = FALSE;
                CPutStr("DESCENT MODE ACTIVATED", 243);
            }
            Init_Interface();
        }

        if (TestKey(PLUS_KEY)) Enlarge_Screen();
        if (TestKey(MINUS_KEY)) Shrink_Screen();
 
        if (TestKey(F5_KEY)) Set_ScreenMode(320, 200);
        if (TestKey(F6_KEY)) Set_ScreenMode(320, 400);
        if (TestKey(F7_KEY)) Set_ScreenMode(640, 480);
        if (TestKey(F8_KEY)) Set_ScreenMode(800, 600);

        Render_Frame();
        
/*
        for (int i=0; i < 17; i++) {
            engine.screen.buffer[100*20+152+i] = 234;
            engine.screen.buffer[92*320+160+i*320] = 234;
        }
*/
        engine.screen.Show();

        #ifdef AUDIO
            AUpdateAudio();
        #endif

        if (TestKey(TAB_KEY)) {

            #ifdef AUDIO
                APlayVoice(voice1, console_sound);
                ASetVoiceFrequency(voice1, 40000);
            #endif
            int code = Console();
            if (code == QUIT) done = TRUE;
            Init_Interface();
        }
		if (KEY(P_KEY)) {
        	while(!(KEY(P_KEY))) Update_Input();
        }
    }
}

//--- Shut down -------------------------------------------------------------

void Stop(char *message, int code)
{
    Close_Keyboard();
    Remove_Timer();
 
    Free_World();
    Close_ScreenMode();
    Close_Engine();
 
/*
    if (!code) {
        printf("\nStatistics...            \n");
        printf("-\n");
        printf("  Sectors          : %-16d\n", engine.nr_sectors);
        printf("  Portals          : %-16d\n", engine.nr_portals);
        printf("  Polygons         : %-16d\n\n", engine.nr_polygons);
        printf("  Frames Drawn     : %-16d\n", engine.framecounter);
        printf("  Total Time(s)    : %-16.1f\n", (float)(engine.total_ticks) / TICKS_PER_SECOND);
        printf("  Average FPS      : %-16.1f\n", ((float)(engine.framecounter) * TICKS_PER_SECOND) / engine.total_ticks);
        printf("-");
    }
*/
    puts(message);
    exit(code);
}
