//****************************************************************************
//
// File         : screen.cpp
//
// Description  : High level screen mode functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "main.h"
#include "video.h"
#include "fixed.h"

#define NR_SCREENSIZES 5

int             firstset = TRUE;
float           screen_sizes[NR_SCREENSIZES] = {1, 0.8, 0.6, 0.4, 0.2};

//--- Init screenmode system ------------------------------------------------

int Init_Screen()
{
    if (Init_Video()) return(TRUE); else return(FALSE);
}

//--- Close down screen system ----------------------------------------------

void Close_Screen(void)
{
    Close_Video();
}

//--- Set screen mode --------------------------------------------------------

int Set_ScreenMode(int xres, int yres)
{
    int    modenr;

    // Search for an appropriate mode...

    modenr = -1;
    for (int i = 0; i < engine.screen.nr_modes; i++) {

        if ((engine.screen.modelist[i].xres == xres) &&
            (engine.screen.modelist[i].yres == yres)) {

            modenr = i;
            break;
        }
    }
    if (modenr == -1) return(FALSE);       // No mode found
	
    // If neccessary, close down last video mode

    Close_VideoMode();
    if (engine.screen.memory) free(engine.screen.memory);
    engine.screen.flags = engine.screen.size = engine.screen.page = 0;
    engine.screen.memory = 0;

    // Store mode information

    engine.screen.xres = xres;
    engine.screen.yres = yres;
    engine.screen.bytes_per_frame = xres * yres;
    engine.screen.currentmode = &(engine.screen.modelist[modenr]);

    // Allocate memory for ymultab and screenbuffer

    engine.screen.memory = (char *)malloc((engine.screen.bytes_per_frame +
        engine.screen.yres * sizeof(int)));
    if (!engine.screen.memory) return(FALSE);

    engine.screen.ymultab = (int *)(engine.screen.memory + engine.screen.bytes_per_frame);

    Set_VideoMode(modenr);

    engine.screen.videomem = engine.screen.lfb;

    if (!engine.screen.videomem) {
		puts("Video failure");    
        free(engine.screen.memory);
        return(FALSE);
    }

    Init_ModeConstants(engine.screen.currentmode->size);

    if (engine.screen.currentmode->flags & MODEINFO_X11)
        engine.screen.buffer = engine.screen.videomem;
    else engine.screen.buffer = engine.screen.memory;


    if (engine.screen.currentmode->flags & MODEINFO_FLIPABLE) Set_PageFlip(TRUE);
    else Set_PageFlip(FALSE);
    Set_Palette(engine.palette_p);
 
    engine.screen.VClear();
    Wait_Retrace();
    Clear_Buffer();

    engine.screen.flags |= SCREENMODE_SET;
 
    return(TRUE);
}

//--- Close current screenmode ----------------------------------------------

void Close_ScreenMode(void)
{
    Close_VideoMode();
    if (engine.screen.memory) free(engine.screen.memory);
    Wait_Retrace();
}

//--- Shrink screen window --------------------------------------------------

void Shrink_Screen(void)
{
    Wait_Retrace();
    engine.screen.VClear();
    if (engine.screen.currentmode->size < 4) {
        engine.screen.currentmode->size++;
        Init_ModeConstants(engine.screen.currentmode->size);
    }
}

//--- Enlarge screen window -------------------------------------------------

void Enlarge_Screen(void)
{
    Wait_Retrace();
    engine.screen.VClear();
    if (engine.screen.currentmode->size > 0) {
        engine.screen.currentmode->size--;
        Init_ModeConstants(engine.screen.currentmode->size);
    }
}

//--- Initialize the screenmode constants -----------------------------------

void Init_ModeConstants(int size)
{
    engine.screen.width = (int)((float)engine.screen.xres * screen_sizes[size]);
    engine.screen.height = (int)((float)engine.screen.yres * screen_sizes[size]);

    if (engine.screen.currentmode->flags & MODEINFO_PAGEFLIP) {
        engine.screen.top_clip = (engine.screen.yres - engine.screen.height) >> 1;
        engine.screen.left_clip = (engine.screen.xres - engine.screen.width) >> 1;
        engine.screen.bottom_clip = engine.screen.top_clip + engine.screen.height;
        engine.screen.right_clip = engine.screen.left_clip + engine.screen.width;
        engine.screen.bytes_per_row = engine.screen.xres;

        engine.screen.xcenter = (int)((engine.screen.xres - 1) / 2);
        engine.screen.ycenter = (int)((engine.screen.yres - 1) / 2);
    }
    else {
        engine.screen.bytes_per_row = engine.screen.width;
        engine.screen.top_clip = 0;
        engine.screen.left_clip = 0;
        engine.screen.right_clip = engine.screen.width;
        engine.screen.bottom_clip = engine.screen.height;

        engine.screen.xcenter = (int)((engine.screen.width - 1) / 2);
        engine.screen.ycenter = (int)((engine.screen.height - 1) / 2);
    }

    engine.screen.pmulx = engine.screen.width * VIEWER_DISTANCE / 320;
    engine.screen.pmuly = engine.screen.height * VIEWER_DISTANCE / 240;
    engine.screen.pmulx_zclip = engine.screen.pmulx / NEAR_CLIP;
    engine.screen.pmuly_zclip = engine.screen.pmuly / NEAR_CLIP;
    engine.screen.pmul_ratio = engine.screen.pmulx / engine.screen.pmuly;

    for (int i = 0; i < engine.screen.yres; i++)
        engine.screen.ymultab[i] = engine.screen.bytes_per_row * i;
}

//--- If possible, toggle pageflipping on/off --------------------------------

int Set_PageFlip(int toggle)
{
    if (toggle) {
        if (!(engine.screen.currentmode->flags & MODEINFO_FLIPABLE))
            return(FALSE);

        engine.screen.videomem = engine.screen.lfb;
        engine.screen.buffer = engine.screen.lfb + engine.screen.bytes_per_frame;
        engine.screen.page = 0;
        engine.screen.Show = Flip_Buffer;
        engine.screen.currentmode->flags |= MODEINFO_PAGEFLIP;
    }
    else {
        if (!(engine.screen.currentmode->flags & MODEINFO_PAGEFLIP)) return(FALSE);
        engine.screen.videomem = engine.screen.lfb;
        engine.screen.buffer = engine.screen.memory;
        engine.screen.Show = Copy_Buffer;
        engine.screen.currentmode->flags &= ~MODEINFO_PAGEFLIP;
        Set_DisplayStart(0, 0);
    }
    Init_ModeConstants(engine.screen.currentmode->size);
    return(TRUE);
}



//--- Flip linear framebuffer pages -----------------------------------------

void Flip_Buffer(void)
{
    if (engine.screen.currentmode->flags & MODEINFO_WVBL) Wait_Retrace();
    if(engine.screen.page == 0) {
        engine.screen.page = 1;
        Set_DisplayStart(0, engine.screen.yres);
    } else {
        engine.screen.page = 0;
        Set_DisplayStart(0, 0);
    }
    char *tmp = engine.screen.buffer;
    engine.screen.buffer = engine.screen.videomem;
    engine.screen.videomem = tmp;

    if (engine.screen.currentmode->flags & MODEINFO_CLEAR) {
        char *dest = engine.screen.buffer +
            (int)(engine.screen.top_clip) * engine.screen.xres +
            (int)(engine.screen.left_clip);

        for (int i = 0; i < engine.screen.height; i++) {
            Dword_Fill(dest, 0, engine.screen.width >> 2);
            dest += engine.screen.xres;
        }
    }
}

//-- Saves the framebuffer as a TGA file ------------------------------------

int Save_TGA(void)
{
    static          int tganum=0;
    int             i;
    char            tgafile[16] = "scrshot";
    TGA_Header      header;
    FILE            *tga_file;

    header.id_length = 0;
    header.color_map_type = header.image_type = 1;
    header.first_color_map_entry = 0;
    header.color_map_length = 256;
    header.image_width = (short)engine.screen.width;
    header.image_height = (short)engine.screen.height;
    header.image_x_origin = 0;
    header.image_y_origin = 0;

    header.bits_per_pixel = 8;
    header.color_map_entry_size = 24;
    header.image_descriptor_bits = 0;

    sprintf(tgafile, "r3d%d.pcx", tganum);

    tga_file = fopen(tgafile,"wb");
    if (!tga_file) return(FALSE);

    fwrite(&header, sizeof(TGA_Header), 1, tga_file);

    unsigned char tmpc;

    for (i = 0; i < 256; i++) {

        tmpc = (char)(engine.palette_p[i * 3 + 2] << 2);
        fwrite(&tmpc, 1, 1, tga_file);

        tmpc = (char)(engine.palette_p[i * 3 + 1] << 2);
        fwrite(&tmpc, 1, 1, tga_file);

        tmpc = (char)(engine.palette_p[i * 3] << 2);
        fwrite(&tmpc, 1, 1, tga_file);
    }

    for (i = engine.screen.height - 1; i >= 0; i--) {
        fwrite(&(engine.screen.buffer[i * engine.screen.width]), 
            engine.screen.width, 1, tga_file);
    }

    fclose(tga_file);
    return(1);
}


