//****************************************************************************
//
// File        : tmap.cpp
//
// Description : Perspective correct texture mapping code...
//               Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include "main.h"
#include "poly.h"
#include "fixed.h"
#include "fpu.h"

#define SUBDIVIDE 16

/*---------------------------------------------------------------------------
 Draw_SPTextPolygon() - Draws a perspective corrected textured polygon
---------------------------------------------------------------------------*/

void Draw_SPTextPolygon(Vertice_2DI *coord, int points, TPlane *tp)
{
    char           *scroffs, *scrpos;
    int            i, y, ydif, miny, maxy, slen;
    int            lx, rx, ldx, rdx, lcount, rcount;
    int            top_index, bottom_index, ledge, redge;
    int            u1, v1, u2, v2;

    float          da, db, dc, cinv;
    Vector_3D      s;

    int color = tp->timestamp;

    // Find top and bottom points

    miny = maxy = coord[0].y;
    top_index = bottom_index = 0;

    for (i = 1; i < points; i++) {
       if (coord[i].y < miny) {
          miny = coord[i].y;
          top_index = i;
       }
       if (coord[i].y > maxy) {
          maxy = coord[i].y;
          bottom_index = i;
       }
    }

    if (maxy == miny) return;
    miny = F2I_Ceil(miny);
    scroffs = (char *)(engine.screen.buffer + engine.screen.ymultab[miny]);
 
    ledge = redge = top_index;
    lcount = rcount = 0;

    // Loop forever

    while (1) {

        // New left edge?

        if (--lcount < 0) {

            if (ledge == bottom_index) return;
            i = ledge + 1;
            if (i == points) i = 0;
            lcount = F2I_Ceil(coord[i].y - FIXED_ONE) - F2I_Ceil(coord[ledge].y);

            while (lcount < 0) {
                if (i == bottom_index) return;
                ledge = i;
                i++;
                if (i == points) i = 0;
                lcount = F2I_Ceil(coord[i].y - FIXED_ONE) - F2I_Ceil(coord[ledge].y);
            }

            // Calculate edge gradients

            y = (coord[ledge].y + (FIXED_ONE - 1)) & ~FIXED_MASK;
            ydif = coord[i].y - coord[ledge].y;

            if (ydif < (FIXED_ONE / 2)) ydif = (FIXED_ONE / 2);

            ldx = Fixed_Div64(coord[i].x - coord[ledge].x, ydif);
            lx = coord[ledge].x;
            lx += Fixed_Mul64(y - coord[ledge].y, ldx);

            ledge = i;
        }

        // New right edge?

        if (--rcount < 0) {

            if (redge == bottom_index) return;
            i = redge-1;
            if (i < 0) i = points - 1;

            rcount = F2I_Ceil(coord[i].y - FIXED_ONE) - F2I_Ceil(coord[redge].y);

            while (rcount < 0) {
                if (i == bottom_index) return;
                redge = i;
                i--;
                if (i < 0) i = points - 1;
                rcount = F2I_Ceil(coord[i].y - FIXED_ONE) - F2I_Ceil(coord[redge].y);
            }

            // Calculate edge gradients

            y = (coord[redge].y + (FIXED_ONE - 1)) & ~FIXED_MASK;
            ydif = coord[i].y - coord[redge].y;
            if (ydif < (FIXED_ONE / 2)) ydif = (FIXED_ONE / 2);

            rdx = Fixed_Div64(coord[i].x - coord[redge].x, ydif);
            rx = coord[redge].x;
            rx += Fixed_Mul64(y - coord[redge].y, rdx);

            redge = i;
        }

        int slx = lx;
        int srx = rx;

        if (slx < engine.screen.cclip->hclip[miny].sx)
            slx = engine.screen.cclip->hclip[miny].sx;
        if (srx > engine.screen.cclip->hclip[miny].ex)
            srx = engine.screen.cclip->hclip[miny].ex;

        slx = F2I_Ceil(slx);
        srx = F2I_Ceil(srx - FIXED_ONE);

        s.x = (float)slx - engine.screen.xcenter;
        s.y = ((float)miny - engine.screen.ycenter) * engine.screen.pmul_ratio;        s.z = (float)engine.screen.pmulx;

        float a = Dot_Product(&s, &(tp->a));
        float b = Dot_Product(&s, &(tp->b));
        float c = Dot_Product(&s, &(tp->c));

        da = tp->a.x * SUBDIVIDE;
        db = tp->b.x * SUBDIVIDE;
        dc = tp->c.x * SUBDIVIDE;

        slen = srx - slx + 1;

        if (slen > 0) {
            cinv = (float)FIXED_ONE / c;   // First pair of u,v coordinates

            scrpos = (char *)scroffs + slx;

//            u2 = (int)(a * cinv);
//            v2 = (int)(b * cinv);
            Float2Int(&u2, a * cinv);          // u2,v2 * FIXED_ONE
            Float2Int(&v2, b * cinv);

            a += da;                       // Step to next subdivision
            b += db;
            c += dc;

            cinv = (float)FIXED_ONE / c;  // Calculate 1/z for next subdiv

            while (slen >= 0) {
                u1 = u2;                    // End becomes start
                Float2Int(&u2, a * cinv);   // Calculate new set of uv coordinates
//                u2 = (int)(a * cinv);

                int p = SUBDIVIDE;
                slen -= SUBDIVIDE;          // p = nr of pixels to draw in the innerloop
                if (slen < 0) p += slen;

                v1 = v2;
                Float2Int(&v2, b * cinv);
//                v2 = (int)(b * cinv);

                a += da;                      // Step to next subdivision
                b += db;
                c += dc;

                int du = (u2 - u1) >> 4;      // Calculate u,v linear
                int dv = (v2 - v1) >> 4;      // interpolation deltas

                cinv = FIXED_ONE / c;         // Do the div...
                                              // while plotting pixels.
                while (p--) {
                    *scrpos++ = tp->texture_p[(v1 >> (FIXED_POINT - 8) & 0x7f00)
                        + (u1 >> FIXED_POINT & 0x7f)];
//                    *scrpos++ = color;          // This works...
//                    *scrpos++ = color + u1 + v1;        // This doesn't

                    u1 += du;
                    v1 += dv;
                }


            }

        }

        lx += ldx;
        rx += rdx;
        miny++;
        scroffs += engine.screen.bytes_per_row;
    }
}
