//***************************************************************************
//
// File         : video.cpp
//
// Description  : DOS/Watcom low level videomode functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//***************************************************************************

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <conio.h>
#include <i86.h>

#include "main.h"
#include "video.h"
#include "tweaked.h"
#include "vesa2.h"

void    Detect_SVGA(void);
int     Set_SVGAMode(int mode);

char            svga_databuffer[2024];
short           oldmode = 0, modelist[64];
char            *pbp;

int             tweaked_rowlen, tweaked_ysize;
int             mode_rowlen[5] = {80,80,90,90,90};
int             mode_ysize[5] = {240,400,270,360,400};

ModeInfo        standard_dosmodes[6] = {{320, 200, 0, 0, MODEINFO_VGA},
                    {320, 240, 0, 1, MODEINFO_XVGA},
                    {320, 400, 0, 2, MODEINFO_XVGA},
                    {360, 270, 0, 3, MODEINFO_XVGA},
                    {360, 360, 0, 4, MODEINFO_XVGA},
                    {360, 400, 0, 5, MODEINFO_XVGA}};


//--- Init video system ------------------------------------------------------

int Init_Video(void)
{
    union REGS  regs;

    regs.h.ah = 0xf;
    int386(0x10, &regs, &regs);
    oldmode = (short)regs.h.al;

    engine.screen.lfb = NULL;
    engine.screen.vendorname = "UNKNOWN";    // Assume standard 128K VGA card
    engine.screen.cardname = "STANDARD VGA";
    engine.screen.videomemory = 131072;
    
    engine.screen.modelist = (ModeInfo *)malloc(6 * sizeof(ModeInfo));
    if (!engine.screen.modelist) return(FALSE);

    // Copy Standard VGA modes...

    memcpy(engine.screen.modelist, standard_dosmodes, 6 * sizeof(ModeInfo));
    engine.screen.nr_modes = 6;

    Detect_SVGA();

    return(TRUE);
}

//--- Close down video system -----------------------------------------------

void Close_Video()
{
    union REGS  regs;

    if (engine.screen.modelist) free(engine.screen.modelist);

    if (oldmode) {
        regs.w.ax = oldmode;
        int386(0x10, &regs, &regs);
    }
}
                       
//--- Set video mode --------------------------------------------------------

int Set_VideoMode(int mode)
{
    int         i;
    Register    *regs;
    union REGS  cpuregs;

    if (mode == 0) {
        engine.screen.Show = Copy_Buffer;
        engine.screen.BClear = Clear_Buffer;
        engine.screen.VClear = Clear_VGA;

        cpuregs.w.ax = 0x13;
        int386(0x10, &cpuregs, &cpuregs);
        engine.screen.lfb = (char *)(0xa0000);
        return(TRUE);
    }

    if (mode < 6) {

        engine.screen.Show = Copy_XVGA;
        engine.screen.BClear = Clear_Buffer;
        engine.screen.VClear = Clear_XVGA;

        switch(mode) {
            case 1 : regs = Mode_320x240regs; break;
            case 2 : regs = Mode_320x400regs; break;
            case 3 : regs = Mode_360x270regs; break;
            case 4 : regs = Mode_360x360regs; break;
            case 5 : regs = Mode_360x400regs; break;
        }

        tweaked_rowlen = mode_rowlen[mode-1];    // Bytes per row
        tweaked_ysize = mode_ysize[mode-1];
        cpuregs.x.eax = 0x13;
        int386(0x10, &cpuregs, &cpuregs);      // Set normal mode 13h

        Wait_Retrace();
        outp(0x3d4, 0x11);                     // Enable writes to registers 0-7
        i = inp(0x3d5) & 0x7F;
        outp(0x3d4, 0x11);
        outp(0x3d5, i);

        i = 25;
        while (i--) OutReg(*regs++);           // Output VGA registers

        // Clear video memory.

        engine.screen.lfb = (char *)(0xa0000);
        return(TRUE);
    }

    engine.screen.Show = Copy_Buffer;
    engine.screen.BClear = Clear_Buffer;
    engine.screen.VClear = Clear_VGA;

    if (Set_SVGAMode(engine.screen.modelist[mode].code)) return(TRUE);
    else return(FALSE);

}

//--- Look for VESA 2.0 video card ------------------------------------------

void Detect_SVGA(void)
{
    int            modes;
    VBEInfo        *vbeinfo_rmp;
    VBEModeInfo    *vbemode_rmp;
    ModeInfo       *modeinfo;
    union REGS     regs;
    static RMRegs  rmregs;
    unsigned short mem_segment, mem_selector;
    char           *buffer;
    short          *modelist_rmp;

    // Allocate low memory for VESA info block using DPMI interuppt

    regs.x.eax = 0x0100;                      // Function 0x100
    regs.x.ebx = 512 / 16;                    // Allocate 512 bytes
    int386(0x31, &regs, &regs);
    if (regs.x.cflag) return;           // Carry = 1 = DPMI Error

    vbeinfo_rmp = (VBEInfo *)((long)(regs.w.ax) << 4);
    vbemode_rmp = (VBEModeInfo *)((long)(regs.w.ax) << 4);
    mem_segment = regs.w.ax;
    mem_selector = regs.w.dx;

    // Request VESA 2.0 Information

    vbeinfo_rmp->VbeSignature = VBE2;         // VBE2 = 'VBE2'

    regs.x.edi = (long)&rmregs;               // Pointer to real mode registers
    regs.w.ax = 0x300;                        // DPMI function 300
    regs.w.bx = 0x10;                         // Int number 0x10
    regs.w.cx = 0;

    // Registers passed to the real mode interuppt

    rmregs.eax = 0x4f00;                      // Function 0x4f00
    rmregs.es = mem_segment;                  // Segment of VBE info block
    rmregs.edi = 0;                           // Offset of VBE info block

    // Simulate real mode interrupt

    int386(0x31, &regs, &regs);

    if ((!regs.x.cflag) && ((rmregs.eax & 0xff)==0x4f)) {
        if ((vbeinfo_rmp->VbeSignature == VESA) && (vbeinfo_rmp->VbeVersion >= 0x200)) {

            engine.screen.videomemory = vbeinfo_rmp->TotalMemory << 16;

            // Copy strings to protected mode memory

            buffer = svga_databuffer;

            strcpy(buffer, (char *)FLAT_PTR(vbeinfo_rmp->OemVendorNamePtr));
            engine.screen.vendorname = buffer;
            buffer += 1 + strlen(buffer);

            strcpy(buffer, (char *)FLAT_PTR(vbeinfo_rmp->OemProductNamePtr));
            engine.screen.cardname = buffer;
            buffer += 1 + strlen(buffer);

            // Copy video mode list to protected mode memory

            modelist_rmp = (short *)FLAT_PTR(vbeinfo_rmp->VideoModePtr);

            modes = 0;
            while(*modelist_rmp != -1)
                modelist[modes++] = *modelist_rmp++;

            modeinfo = &(engine.screen.modelist[engine.screen.nr_modes]);
            realloc(engine.screen.modelist, (engine.screen.nr_modes + modes) * sizeof(ModeInfo));

            // Add modes to mode list

            for (int i = 0; i < modes; i++) {

                regs.x.edi = (long)&rmregs;   // Address to real mode regs
                regs.w.ax = 0x300;            // DPMI function 300
                regs.w.bx = 0x10;             // Invoke int number 0x10
                regs.w.cx = 0;

                // Registers passed to the real mode interuppt

                rmregs.eax = 0x4f01;          // Function 0x4f01
                rmregs.ecx = modelist[i];     // Mode number
                rmregs.es = mem_segment;      // VBEModeInfo buffer segment
                rmregs.edi = 0;               // Buffer offset

                // Simulate real mode interrupt

                int386(0x31, &regs, &regs);

                // Ok, Now extract useful mode information

                if ((vbemode_rmp->ModeAttributes & VBEMODE_CAPS_LINEAR) &&
                    (vbemode_rmp->BitsPerPixel == 8)) {

                    modeinfo->code = modelist[i];
                    modeinfo->xres = vbemode_rmp->XResolution;
                    modeinfo->yres = vbemode_rmp->YResolution;
                    modeinfo->size = 0;
                    modeinfo->flags = MODEINFO_SVGA;
                    if (engine.screen.videomemory >= (modeinfo->xres * modeinfo->yres * 2))
                        modeinfo->flags |= MODEINFO_FLIPABLE | MODEINFO_PAGEFLIP;
 
                    pbp = (char *)(vbemode_rmp->PhysBasePtr);

                    modeinfo++;
                    engine.screen.nr_modes++;
                }
            }
        }
    }

    // Free low memory

    regs.w.ax = 0x0101;
    regs.w.dx = mem_selector;
    int386(0x31, &regs, &regs);
}

//--- Set VESA 2.0 linear framebuffer mode ----------------------------------

int Set_SVGAMode(int mode)
{
    union REGS      regs;

    regs.w.ax = 0x4f02;                          // Function 0x4f02
    regs.w.bx = (short)(mode | VBE_MODEFLAG_LFB | VBE_MODEFLAG_NOCLEAR);
    int386(0x10, &regs, &regs);
    if (regs.w.ax != 0x4f) {
        Close_VideoMode();
        return(FALSE);
    }

    // Map frame buffer into reachable memory using DPMI function 0x800

    regs.w.cx = (short)((long)pbp & 0xffff);
    regs.w.bx = (short)((long)pbp >> 16);

    regs.w.di = 0xffff;                        // Map 4 Megs
    regs.w.si = 0x3f;

    regs.w.ax = 0x800;
    int386(0x31, &regs, &regs);
    if (regs.x.cflag) return(FALSE);            // Carry = 1 = DPMI Error

    engine.screen.lfb = (char *)((regs.w.bx << 16) + regs.w.cx);

    return(TRUE);
}

//--- Close down current video mode -----------------------------------------

void Close_VideoMode()
{
    union REGS regs;

    if (engine.screen.lfb) {

        // Free LFB mapping

        regs.w.cx = (short)((long)engine.screen.lfb & 0xffff);
        regs.w.bx = (short)((long)engine.screen.lfb >> 16);

        regs.w.ax = 0x801;
        int386(0x31, &regs, &regs);

        engine.screen.lfb = NULL;
    }
}

//--- Copy linear frame buffer to video memory ------------------------------

void Copy_Buffer(void)
{
    char        *source, *dest;

    source = engine.screen.buffer;
    dest = engine.screen.videomem +
        ((engine.screen.yres - engine.screen.height) >> 1) * engine.screen.xres +
        ((engine.screen.xres - engine.screen.width) >> 1);

    int dwords = engine.screen.width >> 2;

    if (engine.screen.currentmode->flags & MODEINFO_WVBL) Wait_Retrace();

    if (engine.screen.currentmode->flags & MODEINFO_CLEAR) {
        for (int i = 0; i < engine.screen.height; i++) {
            Dword_CopyClear(dest, source, 0, dwords);
            source += engine.screen.width;
            dest += engine.screen.xres;
        }
    } else {
        for (int i = 0; i < engine.screen.height; i++) {
            Dword_Copy(dest, source, dwords);
            source += engine.screen.width;
            dest += engine.screen.xres;
        }
    }
}


//--- Clear linear frame buffer ---------------------------------------------

void Clear_Buffer(void)
{
    char *dest = engine.screen.buffer +
        (int)(engine.screen.top_clip) * engine.screen.xres +
        (int)(engine.screen.left_clip);

    for (int i = 0; i < engine.screen.height; i++) {
        Dword_Fill(dest, 0, engine.screen.width >> 2);
        dest += engine.screen.xres;
    }

}

void Clear_VGA(void)
{
    if (engine.screen.currentmode->flags & MODEINFO_FLIPABLE)
        Dword_Fill(engine.screen.lfb, 0, (engine.screen.xres * engine.screen.yres * 2) >> 2);
    else Dword_Fill(engine.screen.videomem, 0, (engine.screen.xres * engine.screen.yres) >> 2);
}


//--- Copy buffer to chained video memory -----------------------------------

void Copy_XVGA(void)
{
    int x = (engine.screen.xres - engine.screen.width) >> 1;
    int y = (engine.screen.yres - engine.screen.height) >> 1;

    Tweaked_CopyBlock(x, y, engine.screen.width,
        engine.screen.height, engine.screen.buffer);

    if (engine.screen.currentmode->flags & MODEINFO_CLEAR)
        Dword_Fill(engine.screen.buffer, 0, (engine.screen.width * engine.screen.height) >> 2);
}

void Clear_XVGA(void)
{
    Tweaked_ClearBlock(0, 0, engine.screen.xres, engine.screen.yres, 0);
}


//--- Wait for vertical retrace ---------------------------------------------

void Wait_Retrace(void)
{
    WaitRetrace_ASM();
}

//--- Set VGA palette -------------------------------------------------------

void Set_Palette(char *palette)
{
    SetPalette_ASM(palette);
}

//--- Sets scanline width ---------------------------------------------------

int Set_ScanWidth(long width)
{
    union REGS regs;

    regs.x.eax=0x4f06;
    regs.x.ebx=0;
    regs.x.ecx=width;
    int386(0x10, &regs, &regs);
    if (regs.w.ax!=0x4f) return(FALSE);
    return(TRUE);
}

//--- Sets display start location -------------------------------------------

int Set_DisplayStart(long x, long y)
{
    union REGS regs;

    regs.x.eax=0x4f07;
    regs.x.ebx=0;
    regs.x.ecx=x;
    regs.x.edx=y;
    int386(0x10, &regs, &regs);
    if (regs.w.ax!=0x4f) return(FALSE);
    return(TRUE);
}

//--- Output VGA register. Modified from tweak16b example source ------------

void OutReg(Register r)
{
    switch (r.port) {
        /* First handle special cases: */

        case ATTRCON_ADDR:
            inp(STATUS_ADDR);           /* reset read/write flip-flop */
            outp(ATTRCON_ADDR, r.index | 0x20);
                                  /* ensure VGA output is enabled */
            outp(ATTRCON_ADDR, r.value);
        break;

        case MISC_ADDR:
        case VGAENABLE_ADDR:
            outp(r.port, r.value);      /* directly to the port */
        break;

        case SEQ_ADDR:
        case GRACON_ADDR:
        case CRTC_ADDR:
        default:                       /* This is the default method: */
            outp(r.port, r.index);      /* index to port           */
            outp(r.port+1, r.value);    /* value to port+1         */
        break;
    }
}

//--- Clear block on chained video memory ------------------------------------

void ClearBlockAsm(int xsize, int ysize, int color, char *dest);

#pragma aux ClearBlockAsm = \
            "push  edx           "\
            "mov   dx, 03c4h     "\
            "mov   ecx, eax      "\
            "mov   eax, 0f02h    "\
            "out   dx, ax        "\
            "mov   bh, bl        "\
            "mov   eax, ebx      "\
            "shl   ebx, 16       "\
            "pop   edx           "\
            "mov   bx, ax        "\
            "mov   eax, ecx      "\
            "mov   esi, tweaked_rowlen"\
            "add   eax, eax      "\
            "sub   esi, eax      "\
            "xor   ah, ah        "\
            "shr   ecx, 1        "\
            "jnc   YLoop         "\
            "mov   ah, 1         "\
         "YLoop:                 "\
            "mov   al, cl        "\
         "XLoop:                 "\
            "mov   [edi], ebx    "\
            "add   edi, 4        "\
            "dec   al            "\
            "jnz   XLoop         "\
            "or    ah, ah        "\
            "jz    Even          "\
            "mov   [edi], bx     "\
            "add   edi, 2        "\
         "Even:                  "\
            "add   edi, esi      "\
            "dec   edx           "\
            "jnz   YLoop         "\
            parm[eax] [edx] [ebx] [edi]\
            modify[eax ebx ecx edx edi esi];


void Tweaked_ClearBlock(int x, int y, int xsize, int ysize, int color)
{
    ClearBlockAsm(xsize >> 3, ysize, color, engine.screen.videomem +
        y * tweaked_rowlen + (x >> 2));
}

//--- Copy block to chained video memory ------------------------------------

void CopyBlockAsm(int xsize, int ysize, char *dest, char *source);

#pragma aux CopyBlockAsm = \
            "mov   ebp, eax      "\
            "mov   ecx, eax      "\
            "shr   eax, 1        "\
            "shl   ebp, 3        "\
            "mov   ch, al        "\
            "sub   ebp, 4        "\
         "YLoop:                 "\
            "xor   cl, cl        "\
         "PLoop:                 "\
            "mov   edx, 03c4h    "\
            "mov   eax, 0102h    "\
            "shl   ah, cl        "\
            "push  edi           "\
            "out   dx, ax        "\
            "push  esi           "\
            "mov   dl, ch        "\
         "XLoop:                 "\
            "mov   ah, [esi+12]  "\
            "mov   al, [esi+8]   "\
            "shl   eax, 16       "\
            "mov   ah, [esi+4]   "\
            "mov   al, [esi]     "\
                                  \
            "mov   [edi], eax    "\
            "add   esi, 16       "\
            "add   edi, 4        "\
            "dec   dl            "\
            "jnz   XLoop         "\
                                  \
            "pop   esi           "\
            "inc   cl            "\
            "inc   esi           "\
            "pop   edi           "\
            "cmp   cl, 4         "\
            "jne   PLoop         "\
                                  \
            "add   edi, tweaked_rowlen "\
            "add   esi, ebp      "\
                                  \
            "dec   ebx           "\
            "jnz   YLoop         "\
            parm[eax] [ebx] [edi] [esi]\
            modify[eax ebx ecx edx edi esi ebp];

void Tweaked_CopyBlock(int x, int y, int xsize, int ysize, char *source)
{
    CopyBlockAsm(xsize >> 3, ysize, engine.screen.videomem +
        y * tweaked_rowlen + (x >> 2), source);
}


