//****************************************************************************
//
// File         : video.cpp
//
// Description  : Linux-X11/GCC low level videomode functions.
//                Written by Lars Andersson (lars_and@algonet.se), 1998
//
//****************************************************************************

#include <stdio.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "main.h"
#include "video.h"

#define NR_STDMODES 7

char		open = FALSE, shm = FALSE;
int		screen_num;
Display 	*disp;
Window 		win;
Visual		*visual;
GC		gc;
static 		XShmSegmentInfo shmi;
XImage 		shm_image;
XImage		*xim = NULL;
Colormap	cmap = 0;

ModeInfo        standard_x11modes[NR_STDMODES] = {{320, 200, 0, 0, MODEINFO_X11},
                    {320, 240, 0, 1, MODEINFO_X11},
		    {360, 270, 0, 2, MODEINFO_X11},
		    {400, 300, 0, 1, MODEINFO_X11},
                    {512, 384, 0, 2, MODEINFO_X11},
                    {640, 480, 0, 3, MODEINFO_X11},
                    {800, 600, 0, 4, MODEINFO_X11}};


//--- Init video system -----------------------------------------------

int Init_Video(void)
{
    engine.screen.modelist = NULL;
    engine.screen.flags = 0;

    disp = XOpenDisplay(NULL);
    if (!disp) {
        puts("Error! Can't open display.");
        return(FALSE);
    }
    screen_num = DefaultScreen(disp);
    visual = DefaultVisual(disp, screen_num);
    if(visual->map_entries != 256) {
        puts("Error! Only 8-bit display supported...");
	return(FALSE);
    }

    if (XShmQueryExtension(disp)) {
        shm = TRUE;
        puts("Using SHM...");
    }
    else puts("SHM not available...");

    engine.screen.vendorname = "UNKNOWN";    
    engine.screen.cardname = "X-SERVER";
    engine.screen.videomemory = 131072;       
  
    engine.screen.modelist = (ModeInfo *)(malloc(NR_STDMODES * sizeof(ModeInfo)));
    if (!engine.screen.modelist) return(FALSE);

    // Copy Standard X modes...

    memcpy(engine.screen.modelist, standard_x11modes, NR_STDMODES * sizeof(ModeInfo));

    engine.screen.nr_modes = NR_STDMODES;

    return(TRUE);
}

//--- Close down video system ------------------------------------------- 

void Close_Video()
{
    if (engine.screen.modelist) free(engine.screen.modelist);    
}

//--- Set X11 mode  -----------------------------------------------------

int Set_VideoMode(int mode)
{
    int		i, xres, yres;	
    XGCValues	stgc;
    XEvent	event;
    XSetWindowAttributes 	attr;

    xres = standard_x11modes[mode].xres;
    yres = standard_x11modes[mode].yres;


    win = XCreateSimpleWindow(disp, DefaultRootWindow(disp), 10, 10, xres, yres, 0, 0, 0);
    XStoreName(disp, win, "Roketz 3D");	
    XMapWindow(disp, win);
 
    gc = XCreateGC(disp, win, 0, &stgc); 
    XSetForeground(disp, gc, BlackPixel(disp, screen_num));
    XSetLineAttributes(disp, gc, 0, LineSolid, CapButt, JoinMiter);
    XSetGraphicsExposures(disp, gc, False);

    attr.event_mask = ExposureMask | KeyPressMask | KeyReleaseMask |
        FocusChangeMask | PointerMotionMask | ButtonPressMask | ButtonReleaseMask;

    XChangeWindowAttributes(disp, win, CWEventMask, &attr);

    cmap = XCreateColormap(disp, RootWindow(disp, screen_num), visual, AllocAll);
    XSetWindowColormap(disp, win, cmap);

    for (;;) {
        XNextEvent (disp, &event);
        if (event.type == Expose) break;
    }
//    puts("A");  
    xim = XGetImage (disp, win, 0, 0, xres, yres, AllPlanes, ZPixmap);
//    printf("xim = %x\n", (int)xim);
//    puts("B");

//    shm = FALSE;
    if (shm) {

        shm_image = *xim;
        shmi.shmid = shmget (IPC_PRIVATE, xim->bytes_per_line*xim->height,
			       IPC_CREAT|0777);
        shmi.shmaddr = (char*)shmat (shmi.shmid, 0, 0);
        shmi.readOnly = FALSE;
        XShmAttach (disp, &shmi);
        shm_image.data = shmi.shmaddr;
        engine.screen.lfb = (char*)shmi.shmaddr;
        shm_image.obdata = (char*)&shmi;
//	shm = TRUE;
       	
    }
    else engine.screen.lfb = (char*)(xim->data);

//    printf("LFB = %x\n", engine.screen.lfb);
    open = TRUE;   
    engine.screen.Show = Copy_Buffer;    
    engine.screen.VClear = Clear_Buffer;
    engine.screen.BClear = Clear_Buffer;
    XFlush(disp);
//    puts("C");
    return(TRUE);	

}


//--- Close down current video mode -----------------------------------------

void Close_VideoMode()
{
    if (open) {
	open = FALSE;
        if (shm) XShmDetach (disp, &shmi);
        XDestroyImage(xim);
        XDestroyWindow(disp, win);
    }  
}


//--- Set 256-color palette -------------------------------------------------

void Set_Palette(char *palette)
{
    int		i;
    XColor 	color;

    for (i = 0; i < 256; i++) {
        color.pixel = i;
        color.red   = palette[i+i+i] << 10;
        color.green = palette[i+i+i+1] << 10;
        color.blue  = palette[i+i+i+2] << 10;
        color.flags = DoRed | DoGreen | DoBlue;
        XStoreColor(disp, cmap, &color);
    }
}

//--- Wait for vertical retrace ---------------------------------------------

void Wait_Retrace(void)
{
}

//--- Copy linear frame buffer to video memory ------------------------------

void Copy_Buffer(void)
{
    char        *source, *dest;

    if (shm) {
        XShmPutImage(disp, win, gc, &shm_image, 0, 0, 0, 0, engine.screen.width, 
            engine.screen.height, FALSE);

    }
    else {
        XPutImage(disp, win, gc, xim, 0, 0, 0, 0, engine.screen.width, 
            engine.screen.height);    
    }
    XFlush(disp); 

    if (engine.screen.currentmode->flags & MODEINFO_CLEAR)
        memset(engine.screen.videomem, 0, engine.screen.width * engine.screen.height); 
}

/*
    source = engine.screen.buffer;
    dest = engine.screen.videomem +
        ((engine.screen.yres - engine.screen.height) >> 1) * engine.screen.xres +
        ((engine.screen.xres - engine.screen.width) >> 1);

    int dwords = engine.screen.width >> 2;

    if (engine.screen.currentmode->flags & MODEINFO_WVBL) Wait_Retrace();

    if (engine.screen.currentmode->flags & MODEINFO_CLEAR) {
        for (int i = 0; i < engine.screen.height; i++) {
            Dword_CopyClear(dest, source, 0, dwords);
            source += engine.screen.width;
            dest += engine.screen.xres;
        }
    } else {
        for (int i = 0; i < engine.screen.height; i++) {
            Dword_Copy(dest, source, dwords);
            source += engine.screen.width;
            dest += engine.screen.xres;
        }
*/




//--- Clear linear frame buffer ---------------------------------------------

void Clear_Buffer(void)
{
    char *dest = engine.screen.buffer +
        (int)(engine.screen.top_clip) * engine.screen.xres +
        (int)(engine.screen.left_clip);

    for (int i = 0; i < engine.screen.height; i++) {
        Dword_Fill(dest, 0, engine.screen.width >> 2);
        dest += engine.screen.xres;
    }

}

int Set_DisplayStart(long x, long y) 
{
    return(TRUE);
}




